"""
Normalization
================

Functions and algorithms to normalize signals
"""
import numpy as np


def minmax(signal: np.ndarray) -> np.ndarray:
    """
    minmax normalize a signal for

    min(signal_) = 0
    max(signal_) = 1

    Usage
    -----
    signal_ = minmax(signal)

    Parameters
    ----------
    signal : np.ndarray
        a signal in vector form to normalize

    Returns
    -------
    signal_ : np.ndarray
        the normalized signal
    """
    signal_ = signal - signal.min()
    signal_ = signal_ / signal_.max()
    return signal_


def maxband(signal: np.ndarray, band_ix: int) -> np.ndarray:
    """
    maxband normalize a signal for

    min(signal_) = 0
    signal_[band_ix] = 1

    Usage
    -----
    signal_ = maxband(signal, band_ix)

    Parameters
    ----------
    signal : np.ndarray
        a signal in vector form to normalize

    band_ix : int
        the index (vector position) of the Raman band to normalize to an intensity of 1.

    Returns
    -------
    signal_ : np.ndarray
        the normalized signal
    """
    signal_ = signal - signal.min()
    signal_ = signal_ / signal_[band_ix]
    return signal_


def snv(signal: np.ndarray) -> np.ndarray:
    """
    snv (Standard Normal Variate) normalize a signal for
    mean(signal_) = 0
    std(signal_) = 1

    Usage
    -----
    signal_ = minmax(signal)

    Parameters
    ----------
    signal : np.ndarray
        a signal in vector form to normalize

    Returns
    -------
    signal_ : np.ndarray
        the normalized signal
    """
    signal_ = signal - signal.mean()
    signal_ = signal_ / signal_.std()
    return signal_


def auc(signal: np.ndarray) -> np.ndarray:
    """
    auc (Area Under Curve) normalize a signal for
    min(signal_) = 0
    sum(signal_) = 1

    Usage
    -----
    signal_ = minmax(signal)

    Parameters
    ----------
    signal : np.ndarray
        a signal in vector form to normalize

    Returns
    -------
    signal_ : np.ndarray
        the normalized signal
    """
    signal_ = signal - signal.min()
    signal_ = signal_ / signal_.sum()
    return signal_



"""
Bubblegif is a module used for the generation of gifs that show how the bubblefill algorithm works.
"""
import os
import shutil

import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import savgol_filter
from PIL import Image

from orpl.baseline_removal import grow_bubble, keep_largest

GIFDIR = "gif"
FONTSIZE = 9
fig = plt.figure()


def savefig(figname=None):
    """
    savefig saves the figure in the 'gif/' directory

    Parameters
    ----------
    figname : _type_, optional
        _description_, by default None
    """
    # get number of files in GIFDIR
    if figname:
        plt.savefig(os.path.join(GIFDIR, figname))
    else:
        nfile = len(os.listdir(GIFDIR))
        plt.savefig(os.path.join(GIFDIR, f"{nfile}.png"))


def plotstep0(spectrum: np.ndarray):
    """
    plotstep0 plots bubblefill step 0 -> input spectrum

    Parameters
    ----------
    spectrum : np.ndarray
        The bubblefill input spectrum
    """
    fig.clear()
    plt.plot(spectrum)
    plt.xlabel("Detector pixel (0 to N)")
    plt.ylabel("Intensity [counts]")
    plt.title("Step 0 : Input Spectrum")
    plt.tight_layout()

    # saving figure
    savefig("step_0.png")


def plotstep1(spectrum: np.ndarray, polyfit: np.ndarray, spectrum_: np.ndarray):
    """
    plotstep1 plots bubblefill step 1 -> after global slope removal

    Parameters
    ----------
    spectrum : np.ndarray
        The spectrum after global slope removal
    """
    fig.clear()
    plt.plot(spectrum, label="$S_0$ : Input Spectrum")
    plt.plot(polyfit, label="$P_0$ : linear slope fit", zorder=0)
    plt.plot(spectrum_, label="$S_0 - P_0$", zorder=0)
    plt.xlabel("Detector pixel (0 to N)")
    plt.ylabel("Intensity [counts]")
    plt.title("Step 1 : Global slope removal")
    plt.legend(fontsize=FONTSIZE, framealpha=1)
    plt.tight_layout()

    # saving figure
    savefig("step_1.png")


def plotstep2(spectrum: np.ndarray):
    """
    plotstep2 plots bubblefill step 2 -> normalization to square aspect ratio

    Parameters
    ----------
    spectrum : np.ndarray
        The normalized spectrum
    """

    fig.clear()
    plt.plot(spectrum)
    plt.xlabel("Detector pixel (0 to N)")
    plt.ylabel("Normalized intensity (0 to N) [au]")
    plt.title("Step 2 : Normalization to square aspect ratio")
    plt.tight_layout()

    savefig("step_2.png")


def plotstep3(spectrum: np.ndarray, baseline: np.ndarray):
    """
    plotstep3 plots bubblefill step 3 -> initial baseline fit

    Parameters
    ----------
    spectrum : np.ndarray
        Normalized spectrum
    baseline : np.ndarray
        initial baseline fit (vector of 0)
    """
    fig.clear()
    plt.plot(spectrum, label="Normalized spectrum", zorder=1)
    plt.plot(baseline, label="Initial baseline fit", zorder=0, color="tab:orange")
    plt.xlabel("Detector pixel (0 to N)")
    plt.ylabel("Normalized intensity (0 to N) [au]")
    plt.title("Step 3 : Baseline fit initialization")
    plt.legend(fontsize=FONTSIZE, framealpha=1)
    plt.tight_layout()

    savefig("step_3.png")


def plotbubbleupdate(
    i,
    spectrum: np.ndarray,
    baseline: np.ndarray,
    bubble: np.ndarray,
    touching_point: int,
    left_bound: int,
    right_bound: int,
    min_bubble_widths: list,
):
    """
    plotbubbleupdate Plots the bubblefill bubble growth loop.

    Parameters
    ----------
    i : _type_
        the number of the current loop itteration
    spectrum : np.ndarray
        the normalized spectrum
    baseline : np.ndarray
        the current baseline fit (of that loop)
    bubble : np.ndarray
        the bubble grown during that itteration
    touching_point : int
        the xaxis location where the bubble made contact
    left_bound : int
        the left bound of the bubble (xaxis position)
    right_bound : int
        the right bound of the bubble (xaxis position)
    min_bubble_widths : list
        the smallest allowed bubble width (exit parameter)
    """
    fig.clear()
    plt.plot(spectrum, label="Normalized spectrum", zorder=10)
    plt.plot(baseline, label="Baseline fit", color="tab:orange", zorder=0)
    plt.plot(
        touching_point,
        spectrum[touching_point],
        "x",
        label="Point of contact",
        color="tab:red",
        zorder=11,
    )
    plt.plot(
        range(left_bound, right_bound),
        bubble,
        label=f"Bubble {i}",
        color="tab:red",
        zorder=5,
    )

    # adding smallest bubble
    theta = np.linspace(0, 2 * np.pi, 100)
    if min_bubble_widths is not int:
        radius = max(min_bubble_widths) / 2
    else:
        radius = min_bubble_widths / 2
    a = radius * np.cos(theta)
    a = a - a.min()
    b = radius * np.sin(theta)
    b = max(spectrum) - b - b.max()
    plt.plot(a, b, color="tab:green", label="Smallest allowed bubble")

    plt.legend(loc="upper right", fontsize=FONTSIZE, framealpha=1)
    plt.ylim([-0.05 * max(spectrum), 1.05 * max(spectrum)])
    plt.xlabel("Detector pixel (0 to N)")
    plt.ylabel("Normalized intensity (0 to N) [au]")
    plt.title(f"Step 4 : Bubble growth loop (i={i})")
    plt.tight_layout()

    savefig(figname=f"loop_{i}.png")


def plotstep4(spectrum: np.ndarray, baseline: np.ndarray):
    """
    plotstep4 plots bubblefill results after bubble growth loop

    Parameters
    ----------
    spectrum : _type_
        spectrum
    baseline : _type_
        baseline fit after bubble growth is complete
    """
    fig.clear()
    plt.plot(spectrum, label="Normalized spectrum")
    plt.plot(baseline, label="Baseline fit", color="tab:orange")

    plt.legend(loc="upper right", fontsize=FONTSIZE, framealpha=1)
    plt.ylim([-0.05 * max(spectrum), 1.05 * max(spectrum)])
    plt.xlabel("Detector pixel (0 to N)")
    plt.ylabel("Normalized intensity (0 to N) [au]")
    plt.title(f"Step 4 : Baseline fit after bubble growth process")

    plt.tight_layout()

    savefig("step_4.png")


def plotstep5(spectrum, baseline):
    """
    plotstep5 plots bubblefill step 5, reversing normalization and square aspect ratio

    Parameters
    ----------
    spectrum : np.ndarray
        de-normalized spectrum
    basleine : np.ndarray
        de-normalized baseline
    """
    fig.clear()
    plt.plot(spectrum, label="Input spectrum")
    plt.plot(baseline, label="Baseline fit", color="tab:orange")

    plt.legend(loc="upper right", fontsize=FONTSIZE, framealpha=1)
    plt.ylim([-0.05 * max(spectrum), 1.05 * max(spectrum)])
    plt.xlabel("Detector pixel (0 to N)")
    plt.ylabel("Intensity [counts]")
    plt.title(f"Step 5 : Reversion of normalization and scaling")

    plt.tight_layout()

    savefig("step_5.png")


def plotstep6(spectrum: np.ndarray, baseline: np.ndarray):
    """
    plotstep6 plots bubblefill step 6, savgol filter for smoothing baseline

    Parameters
    ----------
    spectrum : np.ndarray
        input spectrum
    baseline : np.ndarray
        final baseline fit
    """
    fig.clear()
    plt.plot(spectrum, label="Input spectrum")
    plt.plot(baseline, label="Final baseline fit", color="tab:orange")

    plt.legend(loc="upper right", fontsize=FONTSIZE, framealpha=1)
    plt.ylim([-0.05 * max(spectrum), 1.05 * max(spectrum)])
    plt.xlabel("Detector pixel (0 to N)")
    plt.ylabel("Intensity [counts]")
    plt.title(f"Step 6 : smoothing of baseline fit with Savitzky-Golay filter")

    plt.tight_layout()

    savefig("step_6.png")


def plotstep7(spectrum: np.ndarray, baseline: np.ndarray, raman: np.ndarray):
    """
    plotstep7 plots bubblefill step 7, final baseline removal and raman spectrum

    Parameters
    ----------
    spectrum : np.ndarray
        input spectrum
    baseline : np.ndarray
        final baseline fit
    raman : np.ndarray
        computed raman spectrum
    """
    fig.clear()
    plt.plot(spectrum, label="Input spectrum")
    plt.plot(baseline, label="Final baseline fit", color="tab:orange")
    plt.plot(raman, label="Final raman", color="tab:green")

    plt.legend(loc="upper right", fontsize=FONTSIZE, framealpha=1)
    plt.ylim([-0.05 * max(spectrum), 1.05 * max(spectrum)])
    plt.xlabel("Detector pixel (0 to N)")
    plt.ylabel("Intensity [counts]")
    plt.title("Step 7 : Final Raman and baseline")

    plt.tight_layout()

    savefig("step_7.png")


def bubbleloop(
    spectrum: np.ndarray, baseline: np.ndarray, min_bubble_widths: list
) -> np.ndarray:
    """
    bubbleloop itteratively updates a baseline estimate by growing bubbles under a spectrum.

    Usage
    -----
    baseline = bubbleloop(spectrum, baseline, min_bubble_widths)

    Parameters
    ----------
    spectrum : np.ndarray
        the input spectrum
    baseline : np.ndarray
        the initial baseline should be akin to np.zeros(spectrum.shape)
    min_bubble_widths : list
        the minimum bubble widths to use. Can be an array-like or int.
        if array-like -> must be the same length as spectrum and baseline. Useful to specify
        different bubble sizes based on x-coordinates.
        if int -> will use the same width for all x-coordinates.

    Returns
    -------
    baseline : np.ndarray
        the updated baseline
    """
    # initial range is always 0 -> len(s). aka the whole spectrum
    # bubblecue is a list of bubble x-coordinate span as
    # [[x0, x2]_0, [x0, x2]_1, ... [x0, x2]_n]
    # additional bubble regions are added as the loop runs.
    range_cue = [[0, len(spectrum)]]

    i = 0
    while i < len(range_cue):
        # Bubble parameter from bubblecue
        left_bound, right_bound = range_cue[i]
        i += 1

        if left_bound == right_bound:
            continue

        if isinstance(min_bubble_widths, int):
            min_bubble_width = min_bubble_widths
        else:
            min_bubble_width = min_bubble_widths[(left_bound + right_bound) // 2]

        if left_bound == 0 and right_bound != (len(spectrum)):
            # half bubble right
            alignment = "left"
        elif left_bound != 0 and right_bound == (len(spectrum)):
            alignment = "right"
            # half bubble left
        else:

            # Reached minimum bubble width
            if (right_bound - left_bound) < min_bubble_width:
                continue
            # centered bubble
            alignment = "center"

        # new bubble
        bubble, relative_touching_point = grow_bubble(
            spectrum[left_bound:right_bound], alignment
        )
        touching_point = relative_touching_point + left_bound

        # add bubble to baseline by keeping largest value
        baseline[left_bound:right_bound] = keep_largest(
            baseline[left_bound:right_bound], bubble
        )

        # Plot step
        if i > 0:
            plotbubbleupdate(
                i,
                spectrum,
                baseline,
                bubble,
                touching_point,
                left_bound,
                right_bound,
                min_bubble_widths,
            )

        # Add new bubble(s) to bubblecue
        if touching_point == left_bound:
            range_cue.append([touching_point + 1, right_bound])
        elif touching_point == right_bound:
            range_cue.append([left_bound, touching_point - 1])
        else:
            range_cue.append([left_bound, touching_point])
            range_cue.append([touching_point, right_bound])

    return baseline


def bubblefill(
    spectrum: np.ndarray, min_bubble_widths: list = 50, fit_order: int = 1
) -> (np.ndarray, np.ndarray):
    """
    bubblefill splits a spectrum into it's raman and baseline components.

    Usage
    -----
    raman, baseline = bubblefill(spectrum, bubblewidths, fitorder)

    Parameters
    ----------
    spectrum : np.ndarray
        the input spectrum
    min_bubble_widths: list or int
        is the smallest width allowed for bubbles. Smaller values will
        allow bubbles to penetrate further into peaks resulting
        in a more *aggressive* baseline removal. Larger values are more
        *concervative* and might the computed underestimate baseline.
        use list to specify a minimum width that depends on the
        x-coordinate of the bubble. Make sure len(bubblewidths) = len(spectrum).
        Otherwise if bubblewidths [int], the same width is used for all x-coordinates.
    fit_order : int
        the order of the polynomial fit used to remove the *overall* baseline slope.
        Recommendend value is 1 (for linear slope).
        Higher order will result in Runge's phenomena and
        potentially undesirable and unpredictable effects.
        fitorder = 0 is the same as not removing the overall baseline slope

    Returns
    -------
    raman : np.ndarray
        the spectrum's raman component
    baseline : np.ndarray
        the spectrum's baseline component

    Reference
    ---------
    Guillaume Sheehy 2021-01
    """
    # Setup figure save directory
    if os.path.exists(GIFDIR):
        shutil.rmtree(GIFDIR)
        os.mkdir(GIFDIR)
    else:
        os.mkdir(GIFDIR)

    plotstep0(spectrum)

    xaxis = np.arange(len(spectrum))

    # Remove general slope
    poly_fit = np.poly1d(np.polyfit(xaxis, spectrum, fit_order))(xaxis)
    spectrum_ = spectrum - poly_fit

    plotstep1(spectrum, poly_fit, spectrum_)

    # Normalization
    smin = spectrum_.min()  # value needed to return to the original scaling
    spectrum_ = spectrum_ - smin
    scale = spectrum_.max() / len(spectrum)
    spectrum_ = spectrum_ / scale  # Rescale spectrum to X:Y=1:1 (square aspect ratio)

    plotstep2(spectrum_)

    baseline = np.zeros(spectrum_.shape)

    plotstep3(spectrum_, baseline)

    # Bubble loop (this is the bulk of the algorithm)
    baseline = bubbleloop(spectrum_, baseline, min_bubble_widths)

    plotstep4(spectrum_, baseline)

    # Bringing baseline back in original scale
    baseline = baseline * scale + poly_fit + smin

    plotstep5(spectrum, baseline)

    # Final smoothing of baseline (only if bubblewidth is not a list!!!)
    if isinstance(min_bubble_widths, int):
        baseline = savgol_filter(baseline, 2 * (min_bubble_widths // 4) + 1, 3)

    plotstep6(spectrum, baseline)

    raman = spectrum - baseline

    plotstep7(spectrum, baseline, raman)

    return raman, baseline


def bubblegif(
    spectrum: np.ndarray,
    gif_name: str,
    min_bubble_widths: list = 50,
    fit_order: int = 1,
    gif_duration: int = 20,
    loop_duration_ratio: float = 0.5,
):
    """
    bubblegif bubblegif Generates an animated gif of the bubblefill growth loop process and still frame
    images of the other steps.

    Parameters
    ----------
    spectrum : np.ndarray
        _description_
    gif_name : str
        _description_
    min_bubble_widths : list, optional
        _description_, by default 50
    fit_order : int, optional
        _description_, by default 1
    gif_duration : int, optional
        _description_, by default 10
    loop_duration_ratio : float, optional
        _description_, by default 0.5
    """

    # Create png images
    _, _ = bubblefill(spectrum, min_bubble_widths, fit_order)

    # Creating gif parts
    im_gif_1 = []
    im_gif_2 = []
    im_gif_3 = []
    files = os.listdir("gif")
    files = sorted(files, key=lambda x: int(x.split(".")[0].split("_")[-1]))
    files

    for file in files:
        if file.startswith("loop"):
            im_gif_2.append(Image.open(f"gif/{file}"))
        elif len(im_gif_1) < 4:
            im_gif_1.append(Image.open(f"gif/{file}"))
        else:
            im_gif_3.append(Image.open(f"gif/{file}"))

    long_frame_duration = (
        1000
        * (1 - loop_duration_ratio)
        * gif_duration
        / (len(im_gif_1) + len(im_gif_3))
    )
    quick_frame_duration = 1000 * loop_duration_ratio * gif_duration / len(im_gif_2)

    im_gif_1[0].save(
        f"part1.gif",
        save_all=True,
        append_images=im_gif_1[1:],
        optimize=False,
        duration=long_frame_duration,
    )
    im_gif_2[0].save(
        f"part2.gif",
        save_all=True,
        append_images=im_gif_2[1:],
        optimize=False,
        duration=quick_frame_duration,
    )
    im_gif_3[0].save(
        f"part3.gif",
        save_all=True,
        append_images=im_gif_3[1:],
        optimize=False,
        duration=long_frame_duration,
    )

    # Combining gif parts

    im1 = Image.open("part1.gif")
    im2 = Image.open("part2.gif")
    im3 = Image.open("part3.gif")
    Ims = [im1, im2, im3]

    Ims[0].save(gif_name, save_all=True, append_images=Ims[1:], optimize=True, loop=0)

    # Clean up
    shutil.rmtree("gif")
    for i in range(1, 4):
        os.remove(f"part{i}.gif")



import matplotlib
import matplotlib.pyplot as plt
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from PyQt5 import QtWidgets

matplotlib.use("Qt5Agg")


class MplCanvas(FigureCanvasQTAgg):
    def __init__(self, parent=None, width=5, height=4, dpi=100):
        fig = plt.figure(figsize=(width, height), dpi=dpi)
        self.axes = fig.add_subplot(111)
        super().__init__(fig)


class PlotWidget(QtWidgets.QWidget):
    def __init__(self):
        super().__init__()

        self.canvas = MplCanvas(self, width=5, height=4, dpi=96)
        self.canvas.figure.tight_layout()
        self.toolbar = NavigationToolbar(self.canvas, self)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.canvas)
        layout.addWidget(self.toolbar)

        self.setLayout(layout)

    def createLayout(self):

        layout = QtWidgets.QHBoxLayout()
        layout.addWidget(self)
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        return layout



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'src/orpl/gui/uis/mainWindow.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_mainWindow(object):
    def setupUi(self, mainWindow):
        mainWindow.setObjectName("mainWindow")
        mainWindow.resize(921, 715)
        self.centralwidget = QtWidgets.QWidget(mainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.centralwidget)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.tabWidget = QtWidgets.QTabWidget(self.centralwidget)
        self.tabWidget.setObjectName("tabWidget")
        self.tabFileIO = QtWidgets.QWidget()
        self.tabFileIO.setObjectName("tabFileIO")
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout(self.tabFileIO)
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.label = QtWidgets.QLabel(self.tabFileIO)
        self.label.setObjectName("label")
        self.horizontalLayout_3.addWidget(self.label)
        self.textEditDataDir = QtWidgets.QLineEdit(self.tabFileIO)
        self.textEditDataDir.setReadOnly(True)
        self.textEditDataDir.setObjectName("textEditDataDir")
        self.horizontalLayout_3.addWidget(self.textEditDataDir)
        self.buttonSelectDirectory = QtWidgets.QPushButton(self.tabFileIO)
        self.buttonSelectDirectory.setObjectName("buttonSelectDirectory")
        self.horizontalLayout_3.addWidget(self.buttonSelectDirectory)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.treeViewFiles = QtWidgets.QTreeView(self.tabFileIO)
        self.treeViewFiles.setSelectionMode(QtWidgets.QAbstractItemView.ExtendedSelection)
        self.treeViewFiles.setSortingEnabled(True)
        self.treeViewFiles.setObjectName("treeViewFiles")
        self.treeViewFiles.header().setCascadingSectionResizes(False)
        self.horizontalLayout_7.addWidget(self.treeViewFiles)
        self.plainTextMetadata = QtWidgets.QPlainTextEdit(self.tabFileIO)
        self.plainTextMetadata.setReadOnly(True)
        self.plainTextMetadata.setObjectName("plainTextMetadata")
        self.horizontalLayout_7.addWidget(self.plainTextMetadata)
        self.boxDataSelection = QtWidgets.QGroupBox(self.tabFileIO)
        self.boxDataSelection.setObjectName("boxDataSelection")
        self.horizontalLayout_7.addWidget(self.boxDataSelection)
        self.horizontalLayout_7.setStretch(0, 2)
        self.horizontalLayout_7.setStretch(1, 1)
        self.horizontalLayout_7.setStretch(2, 2)
        self.verticalLayout.addLayout(self.horizontalLayout_7)
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.boxDataPlot = QtWidgets.QGroupBox(self.tabFileIO)
        self.boxDataPlot.setObjectName("boxDataPlot")
        self.horizontalLayout_4.addWidget(self.boxDataPlot)
        self.boxXrefPlot = QtWidgets.QGroupBox(self.tabFileIO)
        self.boxXrefPlot.setObjectName("boxXrefPlot")
        self.horizontalLayout_4.addWidget(self.boxXrefPlot)
        self.boxYrefPlot = QtWidgets.QGroupBox(self.tabFileIO)
        self.boxYrefPlot.setObjectName("boxYrefPlot")
        self.horizontalLayout_4.addWidget(self.boxYrefPlot)
        self.verticalLayout.addLayout(self.horizontalLayout_4)
        self.verticalLayout.setStretch(0, 1)
        self.verticalLayout.setStretch(1, 6)
        self.verticalLayout.setStretch(2, 6)
        self.horizontalLayout_8.addLayout(self.verticalLayout)
        self.groupBox = QtWidgets.QGroupBox(self.tabFileIO)
        self.groupBox.setObjectName("groupBox")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(self.groupBox)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.groupBox_4 = QtWidgets.QGroupBox(self.groupBox)
        self.groupBox_4.setObjectName("groupBox_4")
        self.horizontalLayout_14 = QtWidgets.QHBoxLayout(self.groupBox_4)
        self.horizontalLayout_14.setObjectName("horizontalLayout_14")
        self.buttonSelectSpectra = QtWidgets.QPushButton(self.groupBox_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.buttonSelectSpectra.sizePolicy().hasHeightForWidth())
        self.buttonSelectSpectra.setSizePolicy(sizePolicy)
        self.buttonSelectSpectra.setObjectName("buttonSelectSpectra")
        self.horizontalLayout_14.addWidget(self.buttonSelectSpectra)
        self.buttonDropSpectra = QtWidgets.QPushButton(self.groupBox_4)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.buttonDropSpectra.sizePolicy().hasHeightForWidth())
        self.buttonDropSpectra.setSizePolicy(sizePolicy)
        self.buttonDropSpectra.setText("")
        self.buttonDropSpectra.setObjectName("buttonDropSpectra")
        self.horizontalLayout_14.addWidget(self.buttonDropSpectra)
        self.verticalLayout_2.addWidget(self.groupBox_4)
        self.horizontalLayout_11 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_11.setObjectName("horizontalLayout_11")
        self.verticalLayout_2.addLayout(self.horizontalLayout_11)
        self.groupBox_2 = QtWidgets.QGroupBox(self.groupBox)
        self.groupBox_2.setObjectName("groupBox_2")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.groupBox_2)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.horizontalLayout_10 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_10.setObjectName("horizontalLayout_10")
        self.buttonSelectXref = QtWidgets.QPushButton(self.groupBox_2)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.buttonSelectXref.sizePolicy().hasHeightForWidth())
        self.buttonSelectXref.setSizePolicy(sizePolicy)
        self.buttonSelectXref.setObjectName("buttonSelectXref")
        self.horizontalLayout_10.addWidget(self.buttonSelectXref)
        self.buttonDropXref = QtWidgets.QPushButton(self.groupBox_2)
        self.buttonDropXref.setText("")
        self.buttonDropXref.setObjectName("buttonDropXref")
        self.horizontalLayout_10.addWidget(self.buttonDropXref)
        self.verticalLayout_3.addLayout(self.horizontalLayout_10)
        self.radioButtonTylenol = QtWidgets.QRadioButton(self.groupBox_2)
        self.radioButtonTylenol.setChecked(True)
        self.radioButtonTylenol.setObjectName("radioButtonTylenol")
        self.verticalLayout_3.addWidget(self.radioButtonTylenol)
        self.radioButtonNylon = QtWidgets.QRadioButton(self.groupBox_2)
        self.radioButtonNylon.setChecked(False)
        self.radioButtonNylon.setObjectName("radioButtonNylon")
        self.verticalLayout_3.addWidget(self.radioButtonNylon)
        self.verticalLayout_2.addWidget(self.groupBox_2)
        self.groupBox_3 = QtWidgets.QGroupBox(self.groupBox)
        self.groupBox_3.setObjectName("groupBox_3")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout(self.groupBox_3)
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        self.buttonSelectYref = QtWidgets.QPushButton(self.groupBox_3)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.buttonSelectYref.sizePolicy().hasHeightForWidth())
        self.buttonSelectYref.setSizePolicy(sizePolicy)
        self.buttonSelectYref.setObjectName("buttonSelectYref")
        self.horizontalLayout_9.addWidget(self.buttonSelectYref)
        self.buttonDropYref = QtWidgets.QPushButton(self.groupBox_3)
        self.buttonDropYref.setText("")
        self.buttonDropYref.setObjectName("buttonDropYref")
        self.horizontalLayout_9.addWidget(self.buttonDropYref)
        self.verticalLayout_4.addLayout(self.horizontalLayout_9)
        self.radioButtonNIST = QtWidgets.QRadioButton(self.groupBox_3)
        self.radioButtonNIST.setEnabled(False)
        self.radioButtonNIST.setCheckable(True)
        self.radioButtonNIST.setChecked(True)
        self.radioButtonNIST.setObjectName("radioButtonNIST")
        self.verticalLayout_4.addWidget(self.radioButtonNIST)
        self.verticalLayout_2.addWidget(self.groupBox_3)
        spacerItem = QtWidgets.QSpacerItem(20, 94, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_2.addItem(spacerItem)
        self.horizontalLayout_8.addWidget(self.groupBox)
        self.tabWidget.addTab(self.tabFileIO, "")
        self.tabProcessing = QtWidgets.QWidget()
        self.tabProcessing.setObjectName("tabProcessing")
        self.verticalLayout_12 = QtWidgets.QVBoxLayout(self.tabProcessing)
        self.verticalLayout_12.setObjectName("verticalLayout_12")
        self.horizontalLayout_16 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_16.setObjectName("horizontalLayout_16")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout()
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.processingToolBox = QtWidgets.QToolBox(self.tabProcessing)
        self.processingToolBox.setObjectName("processingToolBox")
        self.pageCRR = QtWidgets.QWidget()
        self.pageCRR.setGeometry(QtCore.QRect(0, 0, 122, 342))
        self.pageCRR.setObjectName("pageCRR")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.pageCRR)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.label_4 = QtWidgets.QLabel(self.pageCRR)
        self.label_4.setObjectName("label_4")
        self.verticalLayout_6.addWidget(self.label_4)
        self.spinBoxLeftCrop = QtWidgets.QSpinBox(self.pageCRR)
        self.spinBoxLeftCrop.setObjectName("spinBoxLeftCrop")
        self.verticalLayout_6.addWidget(self.spinBoxLeftCrop)
        self.label_5 = QtWidgets.QLabel(self.pageCRR)
        self.label_5.setObjectName("label_5")
        self.verticalLayout_6.addWidget(self.label_5)
        self.spinBoxRightCrop = QtWidgets.QSpinBox(self.pageCRR)
        self.spinBoxRightCrop.setObjectName("spinBoxRightCrop")
        self.verticalLayout_6.addWidget(self.spinBoxRightCrop)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_6.addItem(spacerItem1)
        self.processingToolBox.addItem(self.pageCRR, "")
        self.page_3 = QtWidgets.QWidget()
        self.page_3.setGeometry(QtCore.QRect(0, 0, 194, 327))
        self.page_3.setObjectName("page_3")
        self.verticalLayout_10 = QtWidgets.QVBoxLayout(self.page_3)
        self.verticalLayout_10.setObjectName("verticalLayout_10")
        self.checkBoxMultiCRR = QtWidgets.QCheckBox(self.page_3)
        self.checkBoxMultiCRR.setObjectName("checkBoxMultiCRR")
        self.verticalLayout_10.addWidget(self.checkBoxMultiCRR)
        self.label_10 = QtWidgets.QLabel(self.page_3)
        self.label_10.setObjectName("label_10")
        self.verticalLayout_10.addWidget(self.label_10)
        self.spinBoxMCRRWidth = QtWidgets.QSpinBox(self.page_3)
        self.spinBoxMCRRWidth.setObjectName("spinBoxMCRRWidth")
        self.verticalLayout_10.addWidget(self.spinBoxMCRRWidth)
        self.label_11 = QtWidgets.QLabel(self.page_3)
        self.label_11.setObjectName("label_11")
        self.verticalLayout_10.addWidget(self.label_11)
        self.spinBoxMCRRthreshold = QtWidgets.QDoubleSpinBox(self.page_3)
        self.spinBoxMCRRthreshold.setSingleStep(0.01)
        self.spinBoxMCRRthreshold.setObjectName("spinBoxMCRRthreshold")
        self.verticalLayout_10.addWidget(self.spinBoxMCRRthreshold)
        self.line = QtWidgets.QFrame(self.page_3)
        self.line.setFrameShape(QtWidgets.QFrame.HLine)
        self.line.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line.setObjectName("line")
        self.verticalLayout_10.addWidget(self.line)
        self.checkBoxSingleCRR = QtWidgets.QCheckBox(self.page_3)
        self.checkBoxSingleCRR.setObjectName("checkBoxSingleCRR")
        self.verticalLayout_10.addWidget(self.checkBoxSingleCRR)
        self.label_12 = QtWidgets.QLabel(self.page_3)
        self.label_12.setObjectName("label_12")
        self.verticalLayout_10.addWidget(self.label_12)
        self.spinBoxSCRRWidth = QtWidgets.QSpinBox(self.page_3)
        self.spinBoxSCRRWidth.setObjectName("spinBoxSCRRWidth")
        self.verticalLayout_10.addWidget(self.spinBoxSCRRWidth)
        self.label_13 = QtWidgets.QLabel(self.page_3)
        self.label_13.setObjectName("label_13")
        self.verticalLayout_10.addWidget(self.label_13)
        self.spinBoxSCRRstd = QtWidgets.QDoubleSpinBox(self.page_3)
        self.spinBoxSCRRstd.setSingleStep(0.01)
        self.spinBoxSCRRstd.setObjectName("spinBoxSCRRstd")
        self.verticalLayout_10.addWidget(self.spinBoxSCRRstd)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_10.addItem(spacerItem2)
        self.processingToolBox.addItem(self.page_3, "")
        self.page_2 = QtWidgets.QWidget()
        self.page_2.setGeometry(QtCore.QRect(0, 0, 226, 335))
        self.page_2.setObjectName("page_2")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.page_2)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.comboBoxBRAlgorithm = QtWidgets.QComboBox(self.page_2)
        self.comboBoxBRAlgorithm.setObjectName("comboBoxBRAlgorithm")
        self.comboBoxBRAlgorithm.addItem("")
        self.comboBoxBRAlgorithm.addItem("")
        self.comboBoxBRAlgorithm.addItem("")
        self.comboBoxBRAlgorithm.addItem("")
        self.verticalLayout_7.addWidget(self.comboBoxBRAlgorithm)
        self.line_4 = QtWidgets.QFrame(self.page_2)
        self.line_4.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_4.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_4.setObjectName("line_4")
        self.verticalLayout_7.addWidget(self.line_4)
        self.label_8 = QtWidgets.QLabel(self.page_2)
        self.label_8.setObjectName("label_8")
        self.verticalLayout_7.addWidget(self.label_8)
        self.spinBoxPolyOrder = QtWidgets.QSpinBox(self.page_2)
        self.spinBoxPolyOrder.setEnabled(False)
        self.spinBoxPolyOrder.setMaximum(15)
        self.spinBoxPolyOrder.setObjectName("spinBoxPolyOrder")
        self.verticalLayout_7.addWidget(self.spinBoxPolyOrder)
        self.line_3 = QtWidgets.QFrame(self.page_2)
        self.line_3.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_3.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_3.setObjectName("line_3")
        self.verticalLayout_7.addWidget(self.line_3)
        self.label_7 = QtWidgets.QLabel(self.page_2)
        self.label_7.setObjectName("label_7")
        self.verticalLayout_7.addWidget(self.label_7)
        self.spinBoxHWS = QtWidgets.QSpinBox(self.page_2)
        self.spinBoxHWS.setEnabled(False)
        self.spinBoxHWS.setMinimum(1)
        self.spinBoxHWS.setObjectName("spinBoxHWS")
        self.verticalLayout_7.addWidget(self.spinBoxHWS)
        self.line_2 = QtWidgets.QFrame(self.page_2)
        self.line_2.setFrameShape(QtWidgets.QFrame.HLine)
        self.line_2.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.line_2.setObjectName("line_2")
        self.verticalLayout_7.addWidget(self.line_2)
        self.label_6 = QtWidgets.QLabel(self.page_2)
        self.label_6.setObjectName("label_6")
        self.verticalLayout_7.addWidget(self.label_6)
        self.spinBoxBubbleWidth = QtWidgets.QSpinBox(self.page_2)
        self.spinBoxBubbleWidth.setEnabled(False)
        self.spinBoxBubbleWidth.setMinimum(1)
        self.spinBoxBubbleWidth.setObjectName("spinBoxBubbleWidth")
        self.verticalLayout_7.addWidget(self.spinBoxBubbleWidth)
        spacerItem3 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_7.addItem(spacerItem3)
        self.processingToolBox.addItem(self.page_2, "")
        self.page = QtWidgets.QWidget()
        self.page.setGeometry(QtCore.QRect(0, 0, 129, 237))
        self.page.setObjectName("page")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout(self.page)
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.radioButtonNoNorm = QtWidgets.QRadioButton(self.page)
        self.radioButtonNoNorm.setChecked(True)
        self.radioButtonNoNorm.setObjectName("radioButtonNoNorm")
        self.verticalLayout_8.addWidget(self.radioButtonNoNorm)
        self.radioButtonMinMax = QtWidgets.QRadioButton(self.page)
        self.radioButtonMinMax.setObjectName("radioButtonMinMax")
        self.verticalLayout_8.addWidget(self.radioButtonMinMax)
        self.radioButtonAUC = QtWidgets.QRadioButton(self.page)
        self.radioButtonAUC.setObjectName("radioButtonAUC")
        self.verticalLayout_8.addWidget(self.radioButtonAUC)
        self.radioButtonSNV = QtWidgets.QRadioButton(self.page)
        self.radioButtonSNV.setChecked(False)
        self.radioButtonSNV.setObjectName("radioButtonSNV")
        self.verticalLayout_8.addWidget(self.radioButtonSNV)
        self.radioButtonMaxBand = QtWidgets.QRadioButton(self.page)
        self.radioButtonMaxBand.setObjectName("radioButtonMaxBand")
        self.verticalLayout_8.addWidget(self.radioButtonMaxBand)
        self.label_9 = QtWidgets.QLabel(self.page)
        self.label_9.setObjectName("label_9")
        self.verticalLayout_8.addWidget(self.label_9)
        self.spinBoxNormBand = QtWidgets.QDoubleSpinBox(self.page)
        self.spinBoxNormBand.setObjectName("spinBoxNormBand")
        self.verticalLayout_8.addWidget(self.spinBoxNormBand)
        spacerItem4 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout_8.addItem(spacerItem4)
        self.processingToolBox.addItem(self.page, "")
        self.verticalLayout_5.addWidget(self.processingToolBox)
        self.checkBoxAutoUpdate = QtWidgets.QCheckBox(self.tabProcessing)
        self.checkBoxAutoUpdate.setLayoutDirection(QtCore.Qt.LeftToRight)
        self.checkBoxAutoUpdate.setObjectName("checkBoxAutoUpdate")
        self.verticalLayout_5.addWidget(self.checkBoxAutoUpdate)
        self.buttonUpdate = QtWidgets.QPushButton(self.tabProcessing)
        self.buttonUpdate.setObjectName("buttonUpdate")
        self.verticalLayout_5.addWidget(self.buttonUpdate)
        self.horizontalLayout_16.addLayout(self.verticalLayout_5)
        self.verticalLayout_11 = QtWidgets.QVBoxLayout()
        self.verticalLayout_11.setObjectName("verticalLayout_11")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.boxRawSignal = QtWidgets.QGroupBox(self.tabProcessing)
        self.boxRawSignal.setObjectName("boxRawSignal")
        self.horizontalLayout_2.addWidget(self.boxRawSignal)
        self.boxIRFCorrection = QtWidgets.QGroupBox(self.tabProcessing)
        self.boxIRFCorrection.setObjectName("boxIRFCorrection")
        self.horizontalLayout_2.addWidget(self.boxIRFCorrection)
        self.verticalLayout_11.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_15 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_15.setObjectName("horizontalLayout_15")
        self.boxBaselineFit = QtWidgets.QGroupBox(self.tabProcessing)
        self.boxBaselineFit.setObjectName("boxBaselineFit")
        self.horizontalLayout_15.addWidget(self.boxBaselineFit)
        self.boxProcessedSignal = QtWidgets.QGroupBox(self.tabProcessing)
        self.boxProcessedSignal.setObjectName("boxProcessedSignal")
        self.horizontalLayout_15.addWidget(self.boxProcessedSignal)
        self.verticalLayout_11.addLayout(self.horizontalLayout_15)
        self.horizontalLayout_16.addLayout(self.verticalLayout_11)
        self.horizontalLayout_16.setStretch(0, 1)
        self.horizontalLayout_16.setStretch(1, 6)
        self.verticalLayout_12.addLayout(self.horizontalLayout_16)
        self.groupBox_5 = QtWidgets.QGroupBox(self.tabProcessing)
        self.groupBox_5.setObjectName("groupBox_5")
        self.horizontalLayout_17 = QtWidgets.QHBoxLayout(self.groupBox_5)
        self.horizontalLayout_17.setObjectName("horizontalLayout_17")
        self.label_14 = QtWidgets.QLabel(self.groupBox_5)
        self.label_14.setObjectName("label_14")
        self.horizontalLayout_17.addWidget(self.label_14)
        self.textEditExportDir = QtWidgets.QLineEdit(self.groupBox_5)
        self.textEditExportDir.setReadOnly(True)
        self.textEditExportDir.setObjectName("textEditExportDir")
        self.horizontalLayout_17.addWidget(self.textEditExportDir)
        self.buttonSelectExportDir = QtWidgets.QPushButton(self.groupBox_5)
        self.buttonSelectExportDir.setObjectName("buttonSelectExportDir")
        self.horizontalLayout_17.addWidget(self.buttonSelectExportDir)
        self.buttonExportData = QtWidgets.QPushButton(self.groupBox_5)
        self.buttonExportData.setObjectName("buttonExportData")
        self.horizontalLayout_17.addWidget(self.buttonExportData)
        self.verticalLayout_12.addWidget(self.groupBox_5)
        self.tabWidget.addTab(self.tabProcessing, "")
        self.tabProfiles = QtWidgets.QWidget()
        self.tabProfiles.setObjectName("tabProfiles")
        self.horizontalLayout_5 = QtWidgets.QHBoxLayout(self.tabProfiles)
        self.horizontalLayout_5.setObjectName("horizontalLayout_5")
        self.label_2 = QtWidgets.QLabel(self.tabProfiles)
        self.label_2.setObjectName("label_2")
        self.horizontalLayout_5.addWidget(self.label_2)
        self.tabWidget.addTab(self.tabProfiles, "")
        self.tabBatchProcessing = QtWidgets.QWidget()
        self.tabBatchProcessing.setObjectName("tabBatchProcessing")
        self.horizontalLayout_6 = QtWidgets.QHBoxLayout(self.tabBatchProcessing)
        self.horizontalLayout_6.setObjectName("horizontalLayout_6")
        self.label_3 = QtWidgets.QLabel(self.tabBatchProcessing)
        self.label_3.setObjectName("label_3")
        self.horizontalLayout_6.addWidget(self.label_3)
        self.tabWidget.addTab(self.tabBatchProcessing, "")
        self.tabLog = QtWidgets.QWidget()
        self.tabLog.setObjectName("tabLog")
        self.verticalLayout_9 = QtWidgets.QVBoxLayout(self.tabLog)
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.horizontalLayout_13 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_13.setObjectName("horizontalLayout_13")
        self.labelLogPath = QtWidgets.QLabel(self.tabLog)
        self.labelLogPath.setObjectName("labelLogPath")
        self.horizontalLayout_13.addWidget(self.labelLogPath)
        spacerItem5 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_13.addItem(spacerItem5)
        self.verticalLayout_9.addLayout(self.horizontalLayout_13)
        self.plainTextLog = QtWidgets.QPlainTextEdit(self.tabLog)
        self.plainTextLog.setEnabled(True)
        self.plainTextLog.setReadOnly(True)
        self.plainTextLog.setObjectName("plainTextLog")
        self.verticalLayout_9.addWidget(self.plainTextLog)
        self.horizontalLayout_12 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_12.setObjectName("horizontalLayout_12")
        spacerItem6 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_12.addItem(spacerItem6)
        self.buttonCopyLog = QtWidgets.QPushButton(self.tabLog)
        self.buttonCopyLog.setObjectName("buttonCopyLog")
        self.horizontalLayout_12.addWidget(self.buttonCopyLog)
        self.verticalLayout_9.addLayout(self.horizontalLayout_12)
        self.tabWidget.addTab(self.tabLog, "")
        self.horizontalLayout.addWidget(self.tabWidget)
        mainWindow.setCentralWidget(self.centralwidget)
        self.statusBar = QtWidgets.QStatusBar(mainWindow)
        self.statusBar.setObjectName("statusBar")
        mainWindow.setStatusBar(self.statusBar)

        self.retranslateUi(mainWindow)
        self.tabWidget.setCurrentIndex(0)
        self.processingToolBox.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(mainWindow)

    def retranslateUi(self, mainWindow):
        _translate = QtCore.QCoreApplication.translate
        mainWindow.setWindowTitle(_translate("mainWindow", "ORPL - GUI"))
        self.label.setText(_translate("mainWindow", "Data Directory:"))
        self.buttonSelectDirectory.setText(_translate("mainWindow", "Select Directory"))
        self.boxDataSelection.setTitle(_translate("mainWindow", "Selected Spectrum"))
        self.boxDataPlot.setTitle(_translate("mainWindow", "Loaded Data"))
        self.boxXrefPlot.setTitle(_translate("mainWindow", "X Reference"))
        self.boxYrefPlot.setTitle(_translate("mainWindow", "Y Reference"))
        self.groupBox.setTitle(_translate("mainWindow", "File selection"))
        self.groupBox_4.setTitle(_translate("mainWindow", "Spectra"))
        self.buttonSelectSpectra.setText(_translate("mainWindow", "Select Spectra"))
        self.groupBox_2.setTitle(_translate("mainWindow", "X-axis Calibration"))
        self.buttonSelectXref.setText(_translate("mainWindow", "Select X-reference"))
        self.radioButtonTylenol.setText(_translate("mainWindow", "Tylenol"))
        self.radioButtonNylon.setText(_translate("mainWindow", "Nylon"))
        self.groupBox_3.setTitle(_translate("mainWindow", "Y-axis Calibration"))
        self.buttonSelectYref.setText(_translate("mainWindow", "Select Y-reference"))
        self.radioButtonNIST.setText(_translate("mainWindow", "NIST SRM-2241"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabFileIO), _translate("mainWindow", "File IO"))
        self.label_4.setText(_translate("mainWindow", "Crop from left"))
        self.label_5.setText(_translate("mainWindow", "Crop from right"))
        self.processingToolBox.setItemText(self.processingToolBox.indexOf(self.pageCRR), _translate("mainWindow", "Xaxis range"))
        self.checkBoxMultiCRR.setText(_translate("mainWindow", "Multi-accumulation Filter"))
        self.label_10.setText(_translate("mainWindow", "Width"))
        self.label_11.setText(_translate("mainWindow", "Disparity Threshold"))
        self.checkBoxSingleCRR.setText(_translate("mainWindow", "Single-accumulation filter"))
        self.label_12.setText(_translate("mainWindow", "Width"))
        self.label_13.setText(_translate("mainWindow", "STD factor"))
        self.processingToolBox.setItemText(self.processingToolBox.indexOf(self.page_3), _translate("mainWindow", "CRR"))
        self.comboBoxBRAlgorithm.setItemText(0, _translate("mainWindow", "None"))
        self.comboBoxBRAlgorithm.setItemText(1, _translate("mainWindow", "IModPoly"))
        self.comboBoxBRAlgorithm.setItemText(2, _translate("mainWindow", "MorphBR"))
        self.comboBoxBRAlgorithm.setItemText(3, _translate("mainWindow", "BubbleFill"))
        self.label_8.setText(_translate("mainWindow", "Polynomial Order (IModPoly)"))
        self.label_7.setText(_translate("mainWindow", "half width window size (MorphBR)"))
        self.label_6.setText(_translate("mainWindow", "Minimum Bubble Width (BubbleFill)"))
        self.processingToolBox.setItemText(self.processingToolBox.indexOf(self.page_2), _translate("mainWindow", "Baseline Removal"))
        self.radioButtonNoNorm.setText(_translate("mainWindow", "None"))
        self.radioButtonMinMax.setText(_translate("mainWindow", "MinMax"))
        self.radioButtonAUC.setText(_translate("mainWindow", "AUC"))
        self.radioButtonSNV.setText(_translate("mainWindow", "SNV"))
        self.radioButtonMaxBand.setText(_translate("mainWindow", "Max band"))
        self.label_9.setText(_translate("mainWindow", "Band selection"))
        self.processingToolBox.setItemText(self.processingToolBox.indexOf(self.page), _translate("mainWindow", "Normalization"))
        self.checkBoxAutoUpdate.setText(_translate("mainWindow", "Auto-update"))
        self.buttonUpdate.setText(_translate("mainWindow", "Update"))
        self.boxRawSignal.setTitle(_translate("mainWindow", "Raw Spectra"))
        self.boxIRFCorrection.setTitle(_translate("mainWindow", "After System Correction"))
        self.boxBaselineFit.setTitle(_translate("mainWindow", "Baseline Fits"))
        self.boxProcessedSignal.setTitle(_translate("mainWindow", "Raman Spectra"))
        self.groupBox_5.setTitle(_translate("mainWindow", "Export Data"))
        self.label_14.setText(_translate("mainWindow", "Export Directory"))
        self.buttonSelectExportDir.setText(_translate("mainWindow", "Select Directory"))
        self.buttonExportData.setText(_translate("mainWindow", "Export Data"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabProcessing), _translate("mainWindow", "Processing"))
        self.label_2.setText(_translate("mainWindow", "Feature to come..."))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabProfiles), _translate("mainWindow", "Profiles"))
        self.label_3.setText(_translate("mainWindow", "Feature to come..."))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabBatchProcessing), _translate("mainWindow", "Batch Processing"))
        self.labelLogPath.setText(_translate("mainWindow", "TextLabel"))
        self.buttonCopyLog.setText(_translate("mainWindow", "Copy Log"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tabLog), _translate("mainWindow", "Log"))



"""
Cosmic Ray
================

Provides Raman spectrum evaluation metrics.
"""
import numpy as np
from orpl.normalization import snv


def raman_snr(
    raman: np.array, baseline: np.array, exposure_time: float, laser_power: float
) -> np.array:
    """
    raman_snr computes the raman signal to noise ratio as described in Dallaire2020.

    snr = raman_snr(raman, baseline, exposure_time, laser_power)

    Parameters
    ----------
    raman : np.array
        An array of raman spectrum. Repeated measurements are along lines, wavelengths are
        along columns
    baseline : np.array
        An array of baseline spectrum. Same configuration as for raman.
    exposure_time : float
    laser_power : float

    Returns
    -------
    np.array
    snr : The computed snr spectrum
    """
    nb_spectrum = raman.shape[0]  # number of spectrum (aka nb of repeated measurements)
    average_raman = raman.mean(axis=1)
    average_baseline = baseline.mean(axis=1)
    snr = (
        np.sqrt(nb_spectrum * exposure_time * laser_power)
        * average_raman
        / np.sqrt(average_raman + average_baseline)
    )
    return snr


def assi(raman: np.ndarray) -> float:
    """
    ASSI computes the Average Signed Squared Intensity of a Raman spectrum.
    The result is bound between -1 and 1 and can be used as a general purpose spectrum
    quality metric/factor.

    Usage
    -----
    quality = ASSI(raman)

    Parameters
    ----------
    raman : np.ndarray
        an input raman spectrum. Normalization has no effect.

    Returns
    -------
    quality : float
        the quality of the raman spectrum as computed by the ASSI metric.
    """
    raman_ = snv(raman)
    deviation_sign = np.sign(raman_)
    deviation2 = (raman_) ** 2
    quality_factor = (deviation_sign * deviation2).mean()
    return quality_factor



"""
ORPL
================
the Open Raman Processing Library is a collection of tools for the processing, analysis and
visualization of Raman spectrum.
"""

from orpl import (
    baseline_removal,
    calibration,
    cosmic_ray,
    datatypes,
    file_io,
    metrics,
    normalization,
    plot,
    synthetic,
)

ORPL_VERSION = "1.0.3"



"""
Tools to create synthetic raman spectra for benchmarking processing algorithms.
"""

import json
from math import log, sqrt
from importlib import resources

# Dependencies
import matplotlib.pyplot as plt
import numpy as np
from scipy.signal import savgol_filter

# loads synthetic presets
with resources.open_text(
    "orpl.data", "synthetic_presets.json", encoding="utf8"
) as preset_path:
    SYNTHETIC_PRESETS = json.load(preset_path)


def gen_raman(
    peak_centers: list,
    peak_heights: list,
    peak_fwhms: list,
    xaxis=None,
    normalize=True,
    plotting=False,
):
    """
    Generates a synthetic Raman spectrum and returns it in a vector (raman_).

        Inputs:
            - x=np.arange(0,1000) [NDARRAY]
                x axis of the spectrum [integers]

            - peak_centers=[265, 500] [list]
                list that contains the raman peak center positions.

            - peak_heights=[0.5, 0.7] [list]
                list that contains the raman peak heights.

            - peak_fwhms=[5, 15] [list]
                list that contains the raman peak full width at half maximum.

            - normalize=True
                if true, the raman spectra will be normalized raman_.max()=1

            - plotting=False
                if true, the function will plot the peaks as they are generated

        Output:

            - raman_ [NDARRAY]
                the total synthetic raman spectrum
    """
    # set default values
    if xaxis is None:
        xaxis = np.arange(1000).astype(float)

    # Generating raman spectrum
    raman_ = np.zeros(xaxis.shape)

    for peak_height, peak_center, fwhm in zip(peak_heights, peak_centers, peak_fwhms):
        c = fwhm / 2 / sqrt(2 * log(2))
        raman_peak = peak_height * np.exp(-((xaxis - peak_center) ** 2) / 2 / c**2)
        raman_ += raman_peak
        if plotting:
            plt.plot(raman_peak)

    if normalize:
        raman_ = raman_ / raman_.max()

    return raman_


def gen_baseline(coefficients, xaxis=None, normalize=True):
    """
    Generates a parametric baseline signal as a polynomial of
    deg=len(coefficients).

        Inputs:

            - coefficients [list]
                list of the coefficients of the polynomial [c0, c1, c2, ...]

            - x=np.arange(1000) [NDARRAY]
                numpy array that contains the x axis used for the polynomial

            - normalize
    """
    # set default values
    if xaxis is None:
        xaxis = np.arange(1000).astype(float)

    # Generating baseline spectrum
    baseline = np.zeros(xaxis.shape)
    for i, coef in enumerate(coefficients):
        if i == 0:
            baseline += coef
        else:
            sign = coef / abs(coef)
            coef = pow(abs(coef), 1 / i)
            baseline += sign * (coef * xaxis) ** i

    if normalize:
        baseline = baseline / baseline.max()
    return baseline


def gen_synthetic_spectrum(
    preset, rb_ratio=1, noise_std=0, baseline_preset=None, normalize=True, xaxis=None
):
    """
    Generates a synthetic raman spectrum from a pure raman, baseline and
    noise signal. The generation uses a specified preset from the presets.json
    file.
        spectrum, raman, baseline, noise = gen_synthetic_spectrum(...)

        Inputs:

            - preset [string]
                a string that contains the preset to be used for the
                generation. Note that the preset MUST be defined in preset.json

                Note: if baseline_preset is not specified [default:None], a
                parametric baseline for the specified preset will be used.
                Do not use this to test baseline removal algorithm!

            - rb_ratio=1 [float]
                Scalar ratio between max(baseline) and max(raman).
                Note that the synthetic spectra are normalized to min(spectrum) = 0
                and max(spectrum) = 1

            - noise_std=0 [float]
                the desired standard deviation of the noise spectrum.
                The noise signal follows a normal distribution.

            - baseline_preset=None [string]
                Experimental baseline signals (WITHOUT RAMAN) are available in
                the preset file. These baselines can be used instead of
                parametric ones to test baseline removal algorithms. Currently,
                the available baseline_presets are ['aluminium', 'nigrosin'].

            - normalize=True [Bool]
                if true, the outputed spectrum will be normalized (max=1)

            - x=None [NDARRAY]
                the desired xaxis [camera pixel] for the generated spectrum. By
                default, x=0,1,2,...,998,999.

        Outputs:

            - spectrum [NDARRAY]
                a synthetic spectrum = baseline + rb_ratio * Raman +
                noise

            - raman [NDARRAY]
                the pure raman spectrum

            - baseline [NDARRAY]
                the pure baseline spectrum

            - noise [NDARRAY]
                the pure noise spectrum

    """
    # set default values
    if xaxis is None:
        xaxis = np.arange(1000).astype(float)

    # check if preset exists
    available_presets = SYNTHETIC_PRESETS.keys()
    if preset not in available_presets:
        raise ValueError(f"preset not available. Choose from {available_presets}")

    # Generation of Raman
    raman = gen_raman(
        peak_centers=SYNTHETIC_PRESETS[preset]["peak_centers"],
        peak_heights=SYNTHETIC_PRESETS[preset]["peak_heights"],
        peak_fwhms=SYNTHETIC_PRESETS[preset]["peak_fwhms"],
    )
    raman = rb_ratio * raman

    # Generation of baseline
    if baseline_preset is None:
        baseline = gen_baseline(
            coefficients=SYNTHETIC_PRESETS[preset]["baseline_coefs"]
        )
    else:
        # check for baseline_preset in presets.json
        available_baselines = SYNTHETIC_PRESETS["baselines"].keys()
        if baseline_preset not in available_baselines:
            raise ValueError(
                f"baseline preset not available. Choose from {available_baselines}"
            )
        # load baseline from presets.json
        baseline = SYNTHETIC_PRESETS["baselines"][baseline_preset]
        # interpolation of baseline to match xaxis
        baseline = np.interp(xaxis, np.arange(len(baseline)).astype(float), baseline)
        # smooth baseline
        baseline = savgol_filter(baseline, window_length=31, polyorder=3)

    # Normalization of baseline
    baseline = baseline - baseline.min()
    baseline = baseline / baseline.max()

    # Generation of noise
    noise = np.random.normal(loc=0, scale=noise_std, size=len(raman))
    # noise = noise - noise.min()

    # Combining everything
    spectrum = baseline + raman + noise

    # Normalization (max=1)
    if normalize:
        scale = 1 / spectrum.max()
        raman = scale * raman
        baseline = scale * baseline
        spectrum = scale * spectrum

    return spectrum, raman, baseline, noise



"""
System Response Correction
================

Provides Raman spectrum instrument/system response correction tools and xaxis calibration.
"""
import itertools
import numpy as np
from scipy.signal import find_peaks

# Unit conversion utilities


def nm2icm(nm: float, nm0: float = 785) -> float:
    """
    nm2icm Converts wavelenghts from nanometers to cm-1 Raman shift

    Usage
    -----
    icm = nm2icm(nm, nm0=785)

    Parameters
    ----------
    nm : float
        wavelength in [nm]
    nm0 : float, optional
        wavelenght of the excitation [nm]

    Returns
    -------
    float
    icm : Raman shift in [cm-1]
    """
    nm = np.array(nm)
    icm = (1 / nm0 - 1 / nm) * 1e7
    return icm


def icm2nm(icm: float, nm0: float = 785) -> float:
    """
    icm2nm Converts an Raman shift [cm-1] into a wavelength [nm].

    Usage
    -----

    nm = icm2nm(icm, nm0=785)

    Parameters
    ----------
    icm : float
        Raman shift in [cm-1]
    nm0 : float, optional
        wavelenght of excitation in [nm], by default 785

    Returns
    -------
    float
    nm : wavelength in [nm]
    """
    icm = np.array(icm)
    nm = 1 / (1 / nm0 - icm * 1e-7)
    return nm


# Xaxis tools


def find_npeaks(
    signal: np.ndarray, ntarget: int, metric: str = "prominence"
) -> np.ndarray:
    """
    find_npeaks finds ntarget peaks in a signal by automatically adjusting
    the threshold of scipy's find_peaks function with a bisection method.

    Usage
    -----
    peak_locations = find_npeaks(signal, ntarget, metric)

    Parameters
    ----------
    spectrum : np.ndarray
        input signal
    ntarget : int
        number of peak to find
    metric : str, optional
        metric used for peak detection, by default "prominence"
        possible metric are ['prominence', 'height']

    Returns
    -------
    peak_locations : np.ndarray
        the index locations of the peaks found in the input signal

    """
    # check if metric is supported
    supported = ["prominence", "height"]
    if metric not in supported:
        raise ValueError("Unsupported metric. Possible metric are " + str(supported))

    # Normalize signal
    spectrum_ = signal / signal.max()
    threshold = 1

    npeakfound = 0
    while npeakfound is not ntarget:

        if metric == "prominence":
            peak_locations, _ = find_peaks(spectrum_, prominence=threshold)
        elif metric == "height":
            peak_locations, _ = find_peaks(spectrum_, height=threshold)

        npeakfound = peak_locations.size

        # Ajust threshold with bisection method
        if npeakfound > ntarget:
            threshold = 1.5 * threshold
        else:
            threshold = 0.5 * threshold

    return peak_locations


def xaxis_from_ref(
    signal: np.ndarray, refx: np.ndarray, refy: np.ndarray, npks: int, deg: int = 2
) -> np.ndarray:
    """
    xaxis_from_ref generates an xaxis for a spectrum based on a reference
    Experimental spectrum peaks are compared with reference spectrum peaks and its
    reference xaxis to generate the xaxis of the experimental spectrum.

    Usage
    -----

    xaxis = xaxis_from_ref(signal, refx, refy, npks, deg)

    Parameters
    ----------
    signal : np.ndarray
        input measured signal
    refx : np.ndarray
        reference signal (y)
    refy : np.ndarray
        reference xaxis
    npks : int
        number of peaks to use. Peaks are identified with find_npeaks()
    deg : int, optional
        the degree of the xaxis polynomial model, by default 2

    Returns
    -------
    xaxis : np.ndarray
        the generated xaxis
    """
    # Normalization of spectrum
    expy = signal / signal.max()
    refy = refy / refy.max()

    # Find peaks
    exp_pid = find_npeaks(expy, npks)
    ref_pid = find_npeaks(refy, npks)

    ref_p = refx[ref_pid]

    coefs = np.polyfit(exp_pid, ref_p, deg=deg)
    xaxis = np.polynomial.polynomial.polyval(np.arange(expy.size), coefs[::-1])
    return xaxis


def xaxis_from_peaks(
    signal: np.ndarray, peaks: np.ndarray, deg: int = 2
) -> (np.ndarray, float):
    """
    xaxis_from_peaks generates an xaxis for a signal based on a list of peak
    positions (nm, cm-1, ...).

    The output is heavily dependent on the find_npeaks peak finder.

    Usage
    -----
    xaxis, residual = xaxis_from_peaks(signal, peaks, deg)

    Parameters
    ----------
    signal : np.ndarray
        input signal
    peaks : np.ndarray
        list of peak positions.
    deg : int, optional
        the degree of the xaxis polynomial model, by default 2, by default 2

    Returns
    -------
    xaxis : np.ndarray
        the generated xaxis

    residual : float
        the polynomial fit residual between the signal's detected peaks and the provided
        peaks.
    """
    # Normalization of spectrum
    expy = signal / signal.max()

    # Find peaks
    npks = len(peaks)
    exp_pid = find_npeaks(expy, npks)

    coefs, residual, _, _, _ = np.polyfit(exp_pid, peaks, deg=deg, full=True)
    residual = float(residual)

    xaxis = np.polynomial.polynomial.polyval(np.arange(expy.size), coefs[::-1])

    return xaxis, residual


def autogenx(signal: np.ndarray, preset: str = "tylenol", deg: int = 2) -> np.ndarray:
    """
    autogenx automatic xaxis generation from preset.

    Usage
    ------
    xaxis = autogenx(spectrum, preset, deg)

    Parameters
    ----------
    signal : np.ndarray
        sample measured spectrum. Raman peaks should be easily visible for
        better results.
    preset : str, optional
        name of a preset to use.
                    choose from ['tylenol', 'nylon'], by default "tylenol"
    deg : int, optional
        _description_, by default 2

    Returns
    -------
    xaxis : np.ndarray
        the generated xaxis [cm-1]
    """
    preset_pks_pos = {
        "tylenol": [797, 858, 1169, 1237, 1324, 1609, 1648],
        "nylon": [956, 1064, 1132, 1235, 1299, 1443, 1632],
    }

    if preset not in preset_pks_pos:
        raise ValueError(f"Invalid preset. Choose from {preset_pks_pos}")

    # Compute xaxis for npeaks combinations in preset peak lists
    # Selecting 5 peaks among the preset's 7 seems good for now, might need to change
    # for something more systematic later... #TODO?
    npeaks = 5
    axises = []
    residuals = []
    for peaks in itertools.combinations(preset_pks_pos[preset], r=npeaks):
        xaxis, residual = xaxis_from_peaks(signal, peaks)
        axises.append(xaxis)
        residuals.append(residual)

    # Get best xaxis from attempts based on best residual
    xaxis = axises[np.argmin(residuals)]

    return xaxis


# Intensity calibrations (y-axis)

# NIST correction
NIST_COEFS = [
    9.71937e-02,
    2.28325e-04,
    -5.86762e-08,
    2.16023e-10,
    -9.77171e-14,
    1.15596e-17,
]


def compute_irf(measured_nist: np.ndarray, xaxis: np.ndarray = None) -> np.ndarray:
    """
    compute_irf computes an Instrument Response Function (IRF) from a measured
    NIST raman spectrum. The IRF can be used to perform a relative intensity
    correction of a measured spectrum;

    corrected_spectrum = measured_spectrum / IRF

    Usage
    ------
        IRF = getCorrectionCurve(measured_nist, xaxis=None)

    Parameters
    ----------
    measured_nist : np.ndarray
        Measured NIST standard spectrum
    xaxis : np.ndarray, optional
        xaxis of the system in cm-1 (should match size of measured_nist).
        If None, the computation will be made in camera pixels and will not be
        accurate as the NIST coefficients are given in cm-1 units.
        by default None

    Returns
    -------
    IRF : np.ndarray
        The system instrument response
    """
    # computing theoretical NIST response for a system's WL range
    if xaxis is None:
        xaxis = np.array(range(0, len(measured_nist)))

    nist_theoretical = np.zeros(xaxis.shape)
    for i, coef in enumerate(NIST_COEFS):
        nist_theoretical += coef * xaxis**i

    # Computing correction curve
    # measured_nist = measured_nist / measured_nist.max()
    instrument_response = measured_nist / nist_theoretical
    instrument_response = instrument_response / instrument_response.max()

    return instrument_response



from dataclasses import dataclass, field
from pathlib import Path
from typing import Type

from numpy import ndarray


@dataclass(frozen=True)
class Metadata:
    # timestamp: int  # epoch [s]
    filepath: Type[Path]
    exposure_time: float  # [ms]
    source_power: float  # [mw]
    details: dict
    comment: str


@dataclass(frozen=True)
class Spectrum:
    metadata: Metadata
    accumulations: ndarray
    background: ndarray
    nbins: int = field(init=False)
    naccumulations: int = field(init=False)
    mean_spectrum: ndarray = field(init=False)

    def __post_init__(self):
        if self.accumulations.ndim > 1:
            object.__setattr__(self, "naccumulations", self.accumulations.shape[1])
            object.__setattr__(self, "mean_spectrum", self.accumulations.mean(axis=1))
        else:
            object.__setattr__(self, "naccumulations", 1)
            object.__setattr__(self, "mean_spectrum", self.accumulations)

        object.__setattr__(self, "nbins", self.accumulations.shape[0])



import logging
import sys
import traceback
from pathlib import Path
from time import strftime
from typing import List, Tuple

import numpy as np
import pyperclip  # module to copy text to clipboard (ctrl + c | ctrl + v)
import qtmodern.styles
from matplotlib.cm import tab10
from PyQt5 import QtGui
from PyQt5.QtWidgets import (
    QApplication,
    QErrorMessage,
    QFileDialog,
    QFileSystemModel,
    QMainWindow,
    QStyle,
)
from ruamel import yaml

from orpl import ORPL_VERSION, file_io
from orpl.baseline_removal import bubblefill, imodpoly, morph_br
from orpl.calibration import autogenx, compute_irf
from orpl.cosmic_ray import crfilter_multi, crfilter_single
from orpl.datatypes import Spectrum
from orpl.file_io import RDF
from orpl.gui.mplcanvas import PlotWidget
from orpl.gui.uis.ui_mainWindow import Ui_mainWindow
from orpl.normalization import auc, maxband, minmax, snv

# Set-up home directory for ORPL
HOME_DIR = Path.home()
ORPL_DIR = HOME_DIR / "orpl"
if not ORPL_DIR.exists():
    ORPL_DIR.mkdir()
LOG_DIR = ORPL_DIR / "logs"
if not LOG_DIR.exists():
    LOG_DIR.mkdir()


# Logging
DEV_MODE = False

# Creates logger
LOG_FORMAT = (
    "%(asctime)s - %(levelname)s"
    "(%(filename)s:%(funcName)s)"
    "(%(filename)s:%(lineno)d) - "
    "%(message)s"
)

if DEV_MODE:
    LOG_PATH = LOG_DIR / "dev.log"
else:
    LOG_PATH = LOG_DIR / f"{strftime('%Y_%m_%d_%H_%M_%S')}.log"


logging.basicConfig(
    filename=LOG_PATH,
    level=logging.DEBUG,
    format=str(LOG_FORMAT),
    filemode="w",
)
logger = logging.getLogger()
# disables matplotlib font msgs
logging.getLogger("matplotlib.font_manager").disabled = True


class qlogHandler(logging.Handler):
    """
    qlogHandler is a custom logger class that links the QPlainTextEdit from the loggin tab
    with the root logger of the application.
    """

    def __init__(self, QPlainTextEdit):
        super().__init__()
        self.loggerTextEdit = QPlainTextEdit
        self.setFormatter(logging.Formatter(LOG_FORMAT))

        # Add current log lines to widget
        # Currently the log is created before it is linked to the widget
        # Not sure if there is a way to get around this, but it works
        with open(LOG_PATH, encoding="utf8") as file:
            log_thus_far = file.read()
        self.loggerTextEdit.setPlainText(log_thus_far.strip())

    def emit(self, record):
        msg = self.format(record)
        self.loggerTextEdit.appendPlainText(msg)
        self.loggerTextEdit.moveCursor(QtGui.QTextCursor.End)


class main_window(Ui_mainWindow, QMainWindow):
    def __init__(self):
        super().__init__()

        # Internal references
        self.raw_spectra: List[np.ndarray] = []
        self.baseline_spectra: List[np.ndarray] = []
        self.irf_corrections: List[np.ndarray] = []
        self.raman_spectra = None
        self.xaxis: Spectrum = None
        self.irf: Spectrum = None
        self.auto_update_processing: bool = False
        self.file_system_model = QFileSystemModel()

        # Plot windows
        self.currentSpectrumPlot = PlotWidget()
        self.loadedDataPlot = PlotWidget()
        self.xrefPlot = PlotWidget()
        self.yrefPlot = PlotWidget()
        self.rawDataPlot = PlotWidget()
        self.processedDataPlot = PlotWidget()
        self.baselineFitPlot = PlotWidget()
        self.irfCorrectionPlot = PlotWidget()

        # Window Setup
        self.setupUi(self)
        self.linkLogger()

        self.defaultSetup()
        self.connectSlots()
        self.setupPlots()

    def linkLogger(self):
        logger.addHandler(qlogHandler(self.plainTextLog))
        logger.setLevel(logging.DEBUG)
        logger.info("linked Logger with logWidget")

    def defaultSetup(self):
        # window setup
        self.setWindowTitle(f"ORPL GUI - {ORPL_VERSION}")

        # File IO tab
        self.file_system_model.setRootPath("/")
        self.treeViewFiles.setModel(self.file_system_model)
        self.treeViewFiles.setRootIndex(self.file_system_model.index(str(HOME_DIR)))
        self.textEditDataDir.setText(str(HOME_DIR))

        pixmapi = getattr(QStyle, "SP_TrashIcon")
        trash_icon = self.style().standardIcon(pixmapi)
        self.buttonDropSpectra.setIcon(trash_icon)
        self.buttonDropXref.setIcon(trash_icon)
        self.buttonDropYref.setIcon(trash_icon)

        # Processing tab
        self.spinBoxMCRRWidth.setValue(3)
        self.spinBoxMCRRthreshold.setValue(0.1)
        self.spinBoxSCRRWidth.setValue(3)
        self.spinBoxSCRRstd.setValue(5)
        self.spinBoxPolyOrder.setValue(6)
        self.spinBoxHWS.setMaximum(1024)
        self.spinBoxHWS.setValue(100)
        self.spinBoxBubbleWidth.setMaximum(1024)
        self.spinBoxBubbleWidth.setValue(200)
        self.textEditExportDir.setText(str(HOME_DIR))

        # Log
        self.labelLogPath.setText(str(LOG_PATH))
        logger.info("setted default logTab setup")

    def setupPlots(self):
        self.boxDataSelection.setLayout(self.currentSpectrumPlot.createLayout())
        self.boxDataPlot.setLayout(self.loadedDataPlot.createLayout())
        self.boxXrefPlot.setLayout(self.xrefPlot.createLayout())
        self.boxYrefPlot.setLayout(self.yrefPlot.createLayout())
        self.boxRawSignal.setLayout(self.rawDataPlot.createLayout())
        self.boxProcessedSignal.setLayout(self.processedDataPlot.createLayout())
        self.boxBaselineFit.setLayout(self.baselineFitPlot.createLayout())
        self.boxIRFCorrection.setLayout(self.irfCorrectionPlot.createLayout())

    def connectSlots(self):
        # File IO tab
        self.buttonSelectDirectory.clicked.connect(self.select_working_directory)
        self.buttonSelectSpectra.clicked.connect(self.select_data)
        self.treeViewFiles.selectionModel().selectionChanged.connect(
            self.selected_file_changed
        )
        self.buttonSelectXref.clicked.connect(self.select_xref)
        self.buttonSelectYref.clicked.connect(self.select_yref)
        self.buttonDropSpectra.clicked.connect(self.drop_data)
        self.buttonDropXref.clicked.connect(self.drop_xref)
        self.buttonDropYref.clicked.connect(self.drop_yref)

        # Processing tab
        self.spinBoxLeftCrop.valueChanged.connect(self.left_crop_changed)
        self.spinBoxRightCrop.valueChanged.connect(self.right_crop_changed)
        self.buttonUpdate.clicked.connect(self.process_spectra)
        self.checkBoxMultiCRR.clicked.connect(self.processing_changed)
        self.spinBoxMCRRWidth.valueChanged.connect(self.processing_changed)
        self.spinBoxMCRRthreshold.valueChanged.connect(self.processing_changed)
        self.checkBoxSingleCRR.clicked.connect(self.processing_changed)
        self.spinBoxSCRRWidth.valueChanged.connect(self.processing_changed)
        self.spinBoxSCRRstd.valueChanged.connect(self.processing_changed)
        self.comboBoxBRAlgorithm.currentTextChanged.connect(self.processing_changed)
        self.spinBoxPolyOrder.valueChanged.connect(self.processing_changed)
        self.spinBoxHWS.valueChanged.connect(self.processing_changed)
        self.spinBoxBubbleWidth.valueChanged.connect(self.processing_changed)
        self.radioButtonNoNorm.clicked.connect(self.processing_changed)
        self.radioButtonMinMax.clicked.connect(self.processing_changed)
        self.radioButtonAUC.clicked.connect(self.processing_changed)
        self.radioButtonSNV.clicked.connect(self.processing_changed)
        self.radioButtonMaxBand.clicked.connect(self.processing_changed)
        self.spinBoxNormBand.valueChanged.connect(self.processing_changed)
        self.buttonSelectExportDir.clicked.connect(self.select_export_dir)
        self.buttonExportData.clicked.connect(self.export_data)

        # self.comboBox

        # Log tab
        self.buttonCopyLog.clicked.connect(self.copy_log)

        logger.info("connected GUI slots")

    # Callbacks

    # Tab File IO

    def selected_file_changed(self):
        logger.info("Updating currentSpectrum plot")
        files = self.get_selected_files()
        if len(files) == 0:
            return

        logger.info("Selected files - %s", files)
        lastfile = files[-1]

        # Load file
        try:
            spectrum = file_io.load_file(lastfile)
        except Exception:
            logger.error(traceback.format_exc())
            return

        # Update metadata
        detail_str = yaml.dump(spectrum.metadata.details, Dumper=yaml.RoundTripDumper)
        self.plainTextMetadata.setPlainText(detail_str)

        # Plot data
        ax = self.currentSpectrumPlot.canvas.axes
        ax.clear()
        ax.plot(spectrum.accumulations)
        ax.set_xlabel("Camera pixel [au]")
        ax.set_ylabel("Intensity [counts]")
        ax.figure.tight_layout()
        ax.figure.canvas.draw()
        self.currentSpectrumPlot.toolbar.update()

    def select_working_directory(self):
        options = QFileDialog.Options()
        new_dir = QFileDialog.getExistingDirectory(
            self, options=options, directory=self.textEditDataDir.text()
        )

        if new_dir:
            self.treeViewFiles.setRootIndex(self.file_system_model.index(new_dir))
            self.textEditDataDir.setText(new_dir)
            self.textEditExportDir.setText(new_dir)

        logger.info("Changed data directory - %s", new_dir)

    def select_data(self):
        # Load selected data
        self.raw_spectra = []
        for file in self.get_selected_files():
            try:
                spectrum = file_io.load_file(file)
                self.raw_spectra.append(spectrum)
                logger.info("Loaded data file - %s", file)
            except Exception:
                logger.error(traceback.format_exc())

        if not self.raw_spectra:
            return

        # Update processing controls
        spectrum_length = self.raw_spectra[0].nbins
        self.spinBoxLeftCrop.setMaximum(spectrum_length)
        self.spinBoxLeftCrop.setValue(0)
        self.spinBoxRightCrop.setMaximum(spectrum_length)
        self.spinBoxRightCrop.setValue(spectrum_length)
        self.spinBoxBubbleWidth.setMaximum(spectrum_length)
        self.spinBoxHWS.setMaximum(spectrum_length)
        self.spinBoxNormBand.setMaximum(spectrum_length)

        # Update plots
        self.plot_loaded_data()
        self.plot_raw_spectra()

    def select_xref(self):
        selected_file = self.get_selected_files()

        if len(selected_file) == 0:
            return
        elif len(selected_file) > 1:
            error_dialog = QErrorMessage()
            error_dialog.showMessage(
                "Please select only ONE file to be used for x-axis calibration."
            )
            error_dialog.exec()
            return

        # Load X-ref
        xref = file_io.load_file(selected_file[0])

        # Compute xaxis from x-ref
        if self.radioButtonTylenol.isChecked():
            xaxis = autogenx(xref.mean_spectrum, preset="tylenol")
        elif self.radioButtonNylon.isChecked():
            xaxis = autogenx(xref.mean_spectrum, preset="nylon")
        self.xaxis = xaxis

        # Update xaxis dependant control
        self.spinBoxNormBand.setMaximum(max(xaxis))

        self.plot_xref(xref)

    def select_yref(self):
        selected_file = self.get_selected_files()

        if len(selected_file) == 0:
            return
        elif len(selected_file) > 1:
            error_dialog = QErrorMessage()
            error_dialog.showMessage(
                "Please select only ONE file to be used for y-axis calibration."
            )
            error_dialog.exec()
            return

        # Load Y-ref
        nist = file_io.load_file(selected_file[0]).mean_spectrum

        # Compute IRF from y-ref
        if self.xaxis is None:
            logger.info("Computing y-ref %s without xaxis", nist.shape)
            irf = compute_irf(nist)
        else:
            logger.info(
                "Computing y-ref %s with xaxis %s", nist.shape, self.xaxis.shape
            )
            irf = compute_irf(nist, xaxis=self.xaxis)

        self.irf = irf

        self.plot_irf()

    def drop_data(self):
        logger.info("Dropping loaded data")
        self.raw_spectra = []

        self.plot_raw_spectra()
        self.plot_loaded_data()

    def drop_xref(self):
        logger.info("Dropping xref and xaxis")
        self.xaxis = None

        self.plot_xref(None)

    def drop_yref(self):
        logger.info("Dropping yref and IRF")
        self.irf = None

        self.plot_irf()

    def plot_loaded_data(self):
        logger.info("Plotting loaded data")
        # Update Spectra graph (in File IO tab)
        ax = self.loadedDataPlot.canvas.axes
        ax.clear()

        for s in self.raw_spectra:
            ax.plot(s.mean_spectrum)
        ax.set_xlabel("Camera pixel [au]")
        ax.set_ylabel("Intensity [counts]")
        ax.figure.tight_layout()
        ax.figure.canvas.draw()
        self.loadedDataPlot.toolbar.update()

    def plot_xref(self, xref: Spectrum):
        logger.info("Plotting (xaxis, xref)")
        # Update X-ref plot
        ax = self.xrefPlot.canvas.axes
        ax.clear()
        if self.xaxis is not None:
            ax.plot(self.xaxis, xref.mean_spectrum)
        ax.set_xlabel(r"Raman Shifts [cm$^{-1}$]")
        ax.set_ylabel("Intensity [counts]")
        ax.figure.tight_layout()
        ax.figure.canvas.draw()
        self.xrefPlot.toolbar.update()

    def plot_irf(self):
        logger.info("Plotting Instrument Response Function")
        # Update Y-ref plot
        ax = self.yrefPlot.canvas.axes
        ax.clear()
        if self.irf is not None:
            if self.xaxis is None:
                ax.plot(self.irf)
            else:
                ax.plot(self.xaxis, self.irf)
        ax.set_xlabel("Camera pixel [au]")
        ax.set_ylabel("Intensity [counts]")
        ax.figure.tight_layout()
        ax.figure.canvas.draw()
        self.yrefPlot.toolbar.update()

    # Tab Processing

    def left_crop_changed(self):
        logger.info("Left crop changed to %s", self.spinBoxLeftCrop.value())

        self.spinBoxRightCrop.setMinimum(self.spinBoxLeftCrop.value() + 10)

        self.plot_raw_spectra()

        if self.checkBoxAutoUpdate.isChecked():
            self.process_spectra()

    def right_crop_changed(self):
        logger.info("Right crop chaneg to %s", self.spinBoxRightCrop.value())

        self.spinBoxLeftCrop.setMaximum(self.spinBoxRightCrop.value() - 10)

        self.plot_raw_spectra()

        if self.checkBoxAutoUpdate.isChecked():
            self.process_spectra()

    def processing_changed(self):
        logger.info("Processing update - %s - changed", self.sender().objectName())

        br_method = self.comboBoxBRAlgorithm.currentText()
        self.spinBoxPolyOrder.setEnabled(False)
        self.spinBoxHWS.setEnabled(False)
        self.spinBoxBubbleWidth.setEnabled(False)
        if br_method == "IModPoly":
            self.spinBoxPolyOrder.setEnabled(True)
        elif br_method == "MorphBR":
            self.spinBoxHWS.setEnabled(True)
        elif br_method == "BubbleFill":
            self.spinBoxBubbleWidth.setEnabled(True)

        if self.checkBoxAutoUpdate.isChecked():
            self.process_spectra()

    def plot_raw_spectra(self):
        logger.info("Updating Raw Spectra plot")
        # Update Spectra graph (in Processing tab)
        ax = self.rawDataPlot.canvas.axes
        ax.clear()

        # Plot raw spectra
        for s in self.raw_spectra:
            ax.plot(s.mean_spectrum)

        # Plot crop lines
        lc = self.spinBoxLeftCrop.value()
        rc = self.spinBoxRightCrop.value()
        ylim = ax.get_ylim()
        ax.plot(
            [lc, lc],
            ylim,
            color="tab:red",
        )
        ax.plot(
            [rc, rc],
            ylim,
            color="tab:red",
        )

        ax.set_xlabel("Camera pixel [au]")
        ax.set_ylabel("Intensity [counts]")
        ax.set_ylim(ylim)
        ax.figure.tight_layout()
        ax.figure.canvas.draw()
        self.rawDataPlot.toolbar.update()

    def process_spectrum(
        self, spectrum: Spectrum
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:

        spectrum_ = spectrum.accumulations
        background_ = spectrum.background

        # Tuncation
        lbound = self.spinBoxLeftCrop.value()
        rbound = self.spinBoxRightCrop.value()
        if spectrum.naccumulations == 1:
            spectrum_ = spectrum_[lbound:rbound]
        else:
            spectrum_ = spectrum_[lbound:rbound, :]
        if background_ is not None:
            background_ = background_[lbound:rbound]

        # CRR filters
        if self.checkBoxMultiCRR.isChecked():
            width = self.spinBoxMCRRWidth.value()
            threshold = self.spinBoxMCRRthreshold.value()
            if spectrum_.ndim > 1:
                spectrum_ = crfilter_multi(
                    spectrum_, width=width, disparity_threshold=threshold
                )
        if self.checkBoxSingleCRR.isChecked():
            width = self.spinBoxSCRRWidth.value()
            std_factor = self.spinBoxSCRRstd.value()
            spectrum_ = crfilter_single(spectrum_, width=width, std_factor=std_factor)

        # Background removal
        if background_ is not None:
            background_ = crfilter_single(background_)
            if background_.ndim != spectrum_.ndim:
                spectrum_ = spectrum_ - np.expand_dims(background_, axis=1)
            else:
                spectrum_ = spectrum_ - background_

        # Combining accumulations
        if spectrum.naccumulations > 1:
            spectrum_ = spectrum_.mean(axis=1)

        # IRF correction
        if self.irf is not None:
            spectrum_ = spectrum_ / self.irf[lbound:rbound]
        irf_correction = spectrum_

        # Baseline removal
        method = self.comboBoxBRAlgorithm.currentText()
        if method == "BubbleFill":
            width = self.spinBoxBubbleWidth.value()
            raman, baseline = bubblefill(spectrum_, min_bubble_widths=width)
        elif method == "IModPoly":
            poly_order = self.spinBoxPolyOrder.value()
            raman, baseline = imodpoly(spectrum_, poly_order=poly_order)
        elif method == "MorphBR":
            hws = self.spinBoxHWS.value()
            raman, baseline = morph_br(spectrum_, hws=hws)
        else:
            raman = spectrum_
            baseline = np.zeros(raman.shape)

        # Normalization
        if self.radioButtonMinMax.isChecked():
            raman = minmax(raman)
        elif self.radioButtonAUC.isChecked():
            raman = auc(raman)
        elif self.radioButtonSNV.isChecked():
            raman = snv(raman)
        elif self.radioButtonMaxBand.isChecked():
            if self.xaxis is not None:
                band_ix = (
                    np.argmin((self.xaxis - self.spinBoxNormBand.value()) ** 2) - lbound
                )
            else:
                band_ix = int(self.spinBoxNormBand.value())
            raman = maxband(raman, band_ix=band_ix)

        return raman, baseline, irf_correction

    def process_spectra(self):
        logger.info("Processing spectra")
        self.raman_spectra = []
        self.baseline_spectra = []
        self.irf_corrections = []
        for s in self.raw_spectra:
            try:
                raman, baseline, irf_correction = self.process_spectrum(s)
                self.raman_spectra.append(raman)
                self.baseline_spectra.append(baseline)
                self.irf_corrections.append(irf_correction)
            except Exception:
                logger.error(traceback.format_exc())

        self.plot_irf_corrections()
        self.plot_baseline_fit()
        self.plot_processed_spectra()

    def plot_processed_spectra(self):
        logger.info("Plotting processed spectra")

        ax = self.processedDataPlot.canvas.axes
        ax.clear()

        for raman in self.raman_spectra:
            if self.xaxis is not None:
                lbound = self.spinBoxLeftCrop.value()
                rbound = self.spinBoxRightCrop.value()
                xaxis = self.xaxis[lbound:rbound]
                ax.plot(xaxis, raman)
                ax.set_xlabel(r"Raman shift [cm$^{-1}$]")
            else:
                ax.plot(raman)
                ax.set_xlabel("Camera pixel")

        if self.radioButtonNoNorm.isChecked():
            ax.set_ylabel("Intensity [counts]")
        else:
            ax.set_ylabel("Normalized Intensity [au]")

        ax.figure.tight_layout()
        ax.figure.canvas.draw()
        self.processedDataPlot.toolbar.update()

    def plot_baseline_fit(self):
        logger.info("Plotting baseline fits")

        ax = self.baselineFitPlot.canvas.axes
        ax.clear()

        for i, (s_before_br, baseline) in enumerate(
            zip(self.irf_corrections, self.baseline_spectra)
        ):
            c = tab10(i % 10)
            lbound = self.spinBoxLeftCrop.value()
            rbound = self.spinBoxRightCrop.value()
            if self.xaxis is not None:
                xaxis = self.xaxis[lbound:rbound]
                ax.plot(xaxis, s_before_br, alpha=0.75, color=c)
                ax.plot(xaxis, baseline, linewidth=1, color=c)
                ax.set_xlabel(r"Raman shift [cm$^{-1}$]")
            else:
                ax.plot(s_before_br, alpha=0.75, color=c)
                ax.plot(baseline, linewidth=1, color=c)
                ax.set_xlabel("Camera pixel")

        ax.set_ylabel("Intensity [counts]")
        ax.figure.tight_layout()
        ax.figure.canvas.draw()
        self.baselineFitPlot.toolbar.update()

    def plot_irf_corrections(self):
        logger.info("Plotting spectra after irf corrections")

        ax = self.irfCorrectionPlot.canvas.axes
        ax.clear()

        for irf_correction in self.irf_corrections:
            ax.plot(irf_correction)

        ax.set_ylabel("Intensity [counts]")
        ax.set_xlabel("Camera pixel [au]")
        ax.figure.tight_layout()
        ax.figure.canvas.draw()
        self.irfCorrectionPlot.toolbar.update()

    def select_export_dir(self):
        logger.info("Selecting new export directory")

        options = QFileDialog.Options()
        new_dir = QFileDialog.getExistingDirectory(
            self, options=options, directory=self.textEditExportDir.text()
        )

        if new_dir:
            self.textEditExportDir.setText(new_dir)

    def get_processing_metadata(self, spectrum: Spectrum) -> dict:
        metadata = {}

        # Software version
        metadata["software_info"] = {"software": "ORPL GUI", "version": ORPL_VERSION}

        # Spectrum info
        metadata["raw_spectrum_path"] = str(spectrum.metadata.filepath)

        # Units
        metadata["units"] = {"xaxis": "camera pixel", "yaxis": "counts"}
        if self.xaxis is not None:
            metadata["units"]["xaxis"] = "cm-1"
        if not self.radioButtonNoNorm.isChecked():
            metadata["units"]["yaxis"] = "au"

        # Truncation
        lbound = self.spinBoxLeftCrop.value()
        rbound = self.spinBoxRightCrop.value()
        metadata["truncation"] = {"left_bound": lbound, "right_bound": rbound}

        # Cosmic ray removal
        metadata["cosmic_ray_removal"] = {
            "single_filter": {"enabled": False},
            "multi_filter": {"enabled": False},
        }
        if self.checkBoxSingleCRR.isChecked():
            width = self.spinBoxSCRRWidth.value()
            std_factor = self.spinBoxSCRRstd.value()
            metadata["cosmic_ray_removal"]["single_filter"]["enabled"] = True
            metadata["cosmic_ray_removal"]["single_filter"]["width"] = width
            metadata["cosmic_ray_removal"]["single_filter"]["std_factor"] = std_factor
        if self.checkBoxMultiCRR.isChecked():
            width = self.spinBoxMCRRWidth.value()
            disparity_threshold = self.spinBoxMCRRthreshold.value()
            metadata["cosmic_ray_removal"]["multi_filter"]["enabled"] = True
            metadata["cosmic_ray_removal"]["multi_filter"]["width"] = width
            metadata["cosmic_ray_removal"]["multi_filter"][
                "disparity_threshold"
            ] = disparity_threshold

        # Instrument response correction
        metadata["instrument_response_correction"] = {"enabled": False}
        if self.irf is not None:
            metadata["instrument_response_correction"]["enabled"] = True
            if self.radioButtonNIST.isChecked():
                metadata["instrument_response_correction"]["method"] = "NIST SRM-2241"

        # Baseline removal
        method = self.comboBoxBRAlgorithm.currentText()
        if method == "BubbleFill":
            width = self.spinBoxBubbleWidth.value()
            metadata["baseline_removal"] = {
                "method": "BubbleFill",
                "min_bubble_width": width,
            }
        elif method == "IModPoly":
            poly_order = self.spinBoxPolyOrder.value()
            metadata["baseline_removal"] = {
                "method": "IModPoly",
                "poly_order": poly_order,
            }
        elif method == "MorphBR":
            poly_order = self.spinBoxPolyOrder.value()
            metadata["baseline_removal"] = {
                "method": "IModPoly",
                "poly_order": poly_order,
            }
        else:
            metadata["baseline_removal"] = {"method": "None"}

        # Normalization
        if self.radioButtonNoNorm.isChecked():
            metadata["normalization"] = {"method": "None"}
        elif self.radioButtonMinMax.isChecked():
            metadata["normalization"] = {"method": "MinMax"}
        elif self.radioButtonAUC.isChecked():
            metadata["normalization"] = {"method": "AUC"}
        elif self.radioButtonSNV.isChecked():
            metadata["normalization"] = {"method": "SNV"}
        elif self.radioButtonMaxBand.isChecked():
            metadata["normalization"] = {
                "method": "MaxBand",
                "band": self.spinBoxNormBand.value(),
            }

        return metadata

    def export_data(self):
        logger.info("Exporting data")

        for i, spectrum in enumerate(self.raw_spectra):
            rdf = RDF()
            rdf.metadata = self.get_processing_metadata(spectrum)

            # Spectra
            rdf.raman = self.raman_spectra[i]
            rdf.baseline = self.baseline_spectra[i]

            # Xaxis
            lbound = self.spinBoxLeftCrop.value()
            rbound = self.spinBoxRightCrop.value()
            if self.xaxis is not None:
                rdf.xaxis = self.xaxis[lbound:rbound]
            else:
                rdf.xaxis = np.arange(lbound, rbound)

            # export file name
            rdf_name = spectrum.metadata.filepath.stem
            export_dir = Path(self.textEditExportDir.text())
            rdf_path = export_dir / (rdf_name + ".rdf")

            # exporting file
            try:
                rdf.save(rdf_path)
            except Exception:
                logger.error(traceback.format_exc())
                continue

            # Logging
            logger.info(f"Exported - {rdf_path}")

    # Tab Log

    def copy_log(self):
        pyperclip.copy(self.plainTextLog.toPlainText())
        logger.info("copied log text")

    # Backend

    def get_selected_files(self) -> list:
        # Get files from selection
        indexes = self.treeViewFiles.selectedIndexes()
        file_names = [index.model().filePath(index) for index in indexes]
        file_names = set(file_names)
        file_paths = [Path(f) for f in file_names]

        # Remove unsupported files
        file_paths = [f for f in file_paths if file_io.is_file_supported(f)]

        return file_paths


def launch_gui():
    logger.info("Starting ORPL GUI")

    app = QApplication(sys.argv)
    # app.setAttribute(Qt.AA_DisableHighDpiScaling, True)

    logger.info("Created application")

    # Changing Theme
    qtmodern.styles.light(app)
    logger.info("Setting Theme")

    window = main_window()
    window.showMaximized()

    app.exec_()


if __name__ == "__main__":

    launch_gui()



"""
file_io modules provide file handling capabilities for the ORPL GUI.
"""

import json
from pathlib import Path
from typing import Type

import numpy as np
import sif_parser
from renishawWiRE import WDFReader
from ruamel import yaml

from orpl.datatypes import Metadata, Spectrum

SUPPORTED_EXTENSIONS = [".sif", ".json", ".wdf"]


def is_file_supported(file_path: Path) -> bool:
    return file_path.suffix in SUPPORTED_EXTENSIONS


def load_file(file_name: str) -> Spectrum:
    file_path = Path(file_name)
    if not is_file_supported(file_path):
        raise TypeError(f"{file_path} is not supported as a spectrum file.")

    load_function = {".sif": load_sif, ".json": load_json, ".wdf": load_wdf}
    spectrum = load_function[file_path.suffix](file_path)

    return spectrum


def load_sif(sif_file: Path) -> Spectrum:
    data_array, meta_dict = sif_parser.np_open(sif_file)
    data_array = np.squeeze(data_array).T
    if data_array.ndim > 1:
        data_array = np.flip(data_array, axis=0)

    # cleaning up string metadata
    for k, v in meta_dict.items():
        if isinstance(v, bytes):
            meta_dict[k] = v.decode(encoding="utf8", errors="replace")
        if isinstance(v, str):
            meta_dict[k] = v.strip()

    metadata = Metadata(
        filepath=sif_file,
        source_power=None,
        exposure_time=meta_dict["CycleTime"],
        details=dict(meta_dict),
        comment="",
    )
    spectrum = Spectrum(accumulations=data_array, metadata=metadata, background=None)

    return spectrum


def load_json(json_file: Path, dump_meta_spectra=True) -> Spectrum:
    with open(json_file, encoding="utf8") as f:
        json_data = json.load(f)

    if isinstance(json_data, list):
        if len(json_data) > 1:
            raise TypeError(
                f"It appears file '{json_file.name}' contains multiple acquisitions. "
                "Please split it into multiple files using the split_json() function."
            )
        else:
            json_data = json_data[0]

    accumulations = np.stack(json_data["RawSpectra"])
    background = np.stack(json_data["Background"])
    details = {
        k: v
        for k, v in json_data.items()
        if k not in ["RawSpectra", "Background", "aec", "asnr"]
    }

    metadata = Metadata(
        filepath=json_file,
        comment=json_data["Comment"],
        exposure_time=json_data["ExpTime"],
        source_power=json_data["Power"],
        details=details,
    )
    spectrum = Spectrum(
        accumulations=accumulations, background=background, metadata=metadata
    )
    return spectrum


def split_json(json_file: Path, new_dir: str = None):
    """
    split_json splits a .json file exported from ORAS into a directory of the same name containing
    a .json file for each indivudual acquisitions.

    Parameters
    ----------
    json_file : str
        the path of a .json file to split into multiple files in a new directory.
    new_dir : str, optional
        the name of the new directory that will contain new individual spectrum files,
        by default None
    """

    # Creating new directory where to split the json_file
    json_path = Path(json_file)
    if json_path.suffix != ".json":
        raise TypeError("json_file is not a .json")

    if new_dir is None:
        new_dir = json_path.resolve().parent / json_path.stem  # New dir Path
    else:
        new_dir = Path(new_dir).resolve()

    if not new_dir.exists():
        new_dir.mkdir()

    # Loading the json_file
    json_path = Path(json_file)

    if not json_path.exists():
        raise FileExistsError(f"{json_path} does not exists.")

    with open(json_path, encoding="utf8") as file:
        json_data = json.load(file)

    if not isinstance(json_data, list):
        raise TypeError("The loaded json does not contain a list.")

    for i, data in enumerate(json_data):
        new_file = new_dir / f"{new_dir.stem}_{i}.json"
        with open(new_file, "w", encoding="utf8") as f:
            json.dump(data, f)


def load_wdf(wdf_path: Path):

    wdf = WDFReader(wdf_path)

    accumulations = np.flip(wdf.spectra)
    details = {
        "software": f"{wdf.application_name} {'.'.join(str(i) for i in wdf.application_version)}",
        "accumulation_count": wdf.accumulation_count,
        "laser_length": wdf.laser_length,
        "measurement_type": str(wdf.measurement_type),
        "point_per_spectrum": wdf.point_per_spectrum,
        "scan_type": str(wdf.scan_type),
        "spectral_unit": str(wdf.spectral_unit),
        "username": wdf.username,
        "xlist_length": wdf.xlist_length,
        "xlist_type": str(wdf.xlist_type),
        "xlist_unit": str(wdf.xlist_unit),
        "ylist_length": wdf.ylist_length,
        "ylist_type": str(wdf.ylist_type),
        "ylist_unit": str(wdf.ylist_unit),
    }

    metadata = Metadata(
        filepath=wdf_path,
        comment=wdf.title,
        exposure_time=None,
        source_power=None,
        details=details,
    )
    spectrum = Spectrum(accumulations=accumulations, background=None, metadata=metadata)
    return spectrum


class SDF:
    pass


class RDF:
    metadata: Type[Metadata]
    xaxis: Type[np.ndarray]
    raman: Type[np.ndarray]
    baseline: Type[np.ndarray]

    def get_metadata_block(self) -> str:
        return "###\n" + yaml.dump(self.metadata, Dumper=yaml.RoundTripDumper) + "###\n"

    def get_data_block(self) -> str:
        data_array = np.stack((self.xaxis, self.baseline, self.raman), axis=1)
        data_str = ""
        for line in data_array:
            data_str += ",".join(map(str, line)) + "\n"
        return data_str

    def get_column_block(self) -> str:
        return "xaxis,baseline,raman\n"

    def save(self, filepath: str) -> None:
        blocks = (
            self.get_metadata_block(),
            self.get_column_block(),
            self.get_data_block(),
        )
        with open(filepath, "w", encoding="utf8") as file:
            for block in blocks:
                file.write(block)


if __name__ == "__main__":
    wdf_file = Path().home() / "Desktop/orpl_testing/nylon_objx5_0.wdf"
    s = load_file(wdf_file)

    print(yaml.dump(s.metadata.details, Dumper=yaml.RoundTripDumper))



"""
Cosmic Ray
================

Provides Raman spectrum cosmic ray removal tools.
"""
import numpy as np
from scipy.ndimage import binary_dilation


def crfilter_single(
    signal: np.ndarray, width: int = 3, std_factor: float = 5
) -> np.ndarray:
    """
    crfilter_single Cosmic Ray filter for a single spectrum/accumulation.

    signal_filtered = crfilter_single(signal, width=3, std_factor=3)

    Parameters
    ----------
    signal : np.ndarray
        The signal to filter.
    width : float, optional
        The filter width, by default 3
    std_factor : float, optional
        The adaptative threshold standard deviation factor.
        Smaller value gives more sensitive filter, by default 3

    Returns
    -------
    signal_filtered : np.ndarray
        The filtered signal.
    """

    # normalize spectrum
    origin = signal.min()
    scale = signal.max()
    signal_normalized = signal - origin
    signal_normalized = signal_normalized / scale

    # find cosmic rays
    # second derivative of spectrum
    diff2 = np.diff(signal_normalized, n=2) ** 2
    threshold = diff2.mean() + std_factor * diff2.std()
    cosmic_ray = np.array([False for i in range(len(signal_normalized))])

    # boolean vector (true if CR)
    cosmic_ray[1:-1] = np.abs(diff2) > threshold

    # include nearest neighbors on each sides
    cosmic_ray = binary_dilation(cosmic_ray, structure=width * [1])

    # removes cosmic rays with linear interpolation
    xaxis = np.arange(len(signal_normalized))
    xaxis_no_cr = xaxis[np.invert(cosmic_ray)]
    spectrum_no_cr = signal_normalized[np.invert(cosmic_ray)]
    signal_filtered = np.interp(xaxis, xaxis_no_cr, spectrum_no_cr)

    # undo normalization
    signal_filtered = scale * signal_filtered
    signal_filtered = signal_filtered + origin

    return signal_filtered


def crfilter_multi(
    signals: np.ndarray, width: int = 3, disparity_threshold: float = 0.1
) -> np.ndarray:
    """
    crfilter_multi Cosmic Ray filter for a multi signal/accumulation acquisition.
    Uses deviation between related signals and find outliers for each wavelengths.
    Filters flagged cosmic rays with linear interpolation.

    signals_filtered = crfilter_multi(signals, width, std_factor)

    Parameters
    ----------
    signals : np.ndarray
        an array of signals (MxN), M are the individual signal and N are wavelengths
    width : float, optional
        filter window width, by default 3
    disparity_threshold : float, optional
        disparity threshold used for detection. Smaller value means
        more sensitive filter. disparity_threshold of 0.1 means a disparity of ~10%
        from singals' mean, by default 0.1

    Returns
    -------
    signals_filtered : np.ndarray
        The filtered signals
    """

    signals_filtered = signals.copy()
    xaxis = np.arange(signals.shape[1])
    mean = signals.mean(0)
    mean_normalized = mean / mean.sum()

    for i, signal in enumerate(signals_filtered):
        # Finding cosmic rays
        disparity = (
            np.abs(signal / signal.sum() - mean_normalized) / mean_normalized.max()
        )
        flagged_cr = disparity > disparity_threshold

        if flagged_cr.sum() == 0:
            # No detected cosmic rays
            continue

        # Widening detection with width parameter
        flagged_cr = binary_dilation(flagged_cr, structure=width * [1])

        # Removing cosmic rays with interpolation
        signal_filtered = np.interp(
            xaxis, xaxis[np.invert(flagged_cr)], signal[np.invert(flagged_cr)]
        )

        # Updating spectrum in spectra_
        signals_filtered[i, :] = signal_filtered

    return signals_filtered


class CRFilter:
    """
    Cosmic ray Filter removal utility.
    """



"""
Plot
================

Provides Raman spectrum visualization tools.
"""
import numpy as np
import matplotlib.pyplot as plt


# Spectrum visualization


def shadeplot(spectra: np.ndarray, labels: np.ndarray, xaxis: np.ndarray = None):
    """
    shadeplot [summary]

    Parameters
    ----------
    spectra : np.ndarray
        [description]
    labels : np.ndarray
        [description]
    xaxis : np.ndarray, optional
        [description], by default None
    """
    if xaxis is None:
        xaxis = np.arange(spectra.shape[1])

    for label in np.unique(labels):
        plt.plot(xaxis, spectra[labels == label].mean(axis=0), label=label)
        bottom_line = spectra[labels == label].mean(axis=0) - spectra[
            labels == label
        ].std(axis=0)
        top_line = spectra[labels == label].mean(axis=0) + spectra[labels == label].std(
            axis=0
        )
        plt.fill_between(xaxis, bottom_line, top_line, alpha=0.4)

    plt.legend()


# Hyperspectral visualization


def plot3d(im3d, ijk=None, xaxis=None, color_map="gist_gray"):
    """
    Visualization tool for hyperspectral cubes.

    Usage
    ------
    plot3d(im3d, ijk=None, xaxis=None, color_map='gist_gray')

    Input arguments
    ----------------
    im3d        --> [NDARRAY] a data cube. MUST BE NxMxP

    ijk         --> [list] the starting (i, j, k) coordinate of the cursor

    xaxis       --> [list] a xaxis. MUST BE 1xP

    color_map   --> [string] colormap to use for imshow. MUST BE A VALID
                    matplotlib COLORMAP NAME
    """
    # Default parameters
    if ijk is None:
        ijk = list(np.array(im3d.shape) // 2)

    if xaxis is None:
        xaxis = np.arange(im3d.shape[2])

    # Create figure
    figure = plt.figure(figsize=[10, 5])

    # Setup left axes
    ax1 = plt.subplot(1, 2, 1)
    ax1.set_title(f"ijk = {ijk}")
    im2d = plt.imshow(im3d[:, :, ijk[2]], cmap=color_map)
    im2d.set_autoscale = True

    hline1 = plt.plot([-1, im3d.shape[1]], [ijk[0], ijk[0]], "red")
    vline1 = plt.plot([ijk[1], ijk[1]], [-1, im3d.shape[0]], "red")
    plt.xlim(-0.5, im3d.shape[1] - 0.5)
    plt.ylim(im3d.shape[0] - 0.5, -0.5)

    # Setup right axes
    ax2 = plt.subplot(1, 2, 2)

    line = plt.plot(xaxis, im3d[ijk[0], ijk[1], :])
    vline2 = plt.plot(
        [xaxis[ijk[2]], xaxis[ijk[2]]], [0.9 * im3d.min(), 1.1 * im3d.max()], "red"
    )

    plt.xlim(xaxis.min(), xaxis.max())
    plt.ylim(0.9 * im3d.min(), 1.1 * im3d.max())
    ax2.set_title(f"({int(round(xaxis[ijk[2]]))}, {im3d[ijk[0], ijk[1], ijk[2]]})")

    def update_axis(ijk):
        # Updates left axes
        hline1[0].set_ydata([ijk[0], ijk[0]])
        vline1[0].set_xdata([ijk[1], ijk[1]])
        im2d.set_data(im3d[:, :, ijk[2]])
        im2d.set_norm(None)

        # Updates right axes
        vline2[0].set_xdata([xaxis[ijk[2]], xaxis[ijk[2]]])
        line[0].set_ydata(im3d[ijk[0], ijk[1]])

        # Update titles
        ax1.set_title(f"ijk = {ijk}")
        ax2.set_title(f"({int(round(xaxis[ijk[2]]))}, {im3d[ijk[0], ijk[1], ijk[2]]})")
        figure.canvas.draw()

    # Setup click behavior

    def onclick(event):
        if event.inaxes == ax1:
            ijk[1] = int(event.xdata + 0.5)
            ijk[1] = max(0, min(ijk[1], im3d.shape[1] - 1))
            ijk[0] = int(event.ydata + 0.5)
            ijk[0] = max(0, min(ijk[0], im3d.shape[0] - 1))

            update_axis(ijk)

        elif event.inaxes == ax2:
            ijk[2] = np.abs(xaxis - int(event.xdata + 0.5)).argmin()
            ijk[2] = max(0, min(ijk[2], im3d.shape[2] - 1))

            update_axis(ijk)

    def keypress(event):
        if event.inaxes == ax2:
            if event.key == "up":
                ijk[2] += 1
            elif event.key == "down":
                ijk[2] -= 1
            elif event.key == "left":
                ijk[2] -= 1
            elif event.key == "right":
                ijk[2] += 1
        else:
            if event.key == "up":
                ijk[0] -= 1
            elif event.key == "down":
                ijk[0] += 1
            elif event.key == "left":
                ijk[1] -= 1
            elif event.key == "right":
                ijk[1] += 1

        ijk[0] = max(0, min(ijk[0], im3d.shape[0] - 1))
        ijk[1] = max(0, min(ijk[1], im3d.shape[1] - 1))
        ijk[2] = max(0, min(ijk[2], im3d.shape[2] - 1))

        update_axis(ijk)

    # Link click behavior to figure
    figure.canvas.mpl_connect("button_press_event", onclick)
    figure.canvas.mpl_connect("key_press_event", keypress)



"""
Baseline Removal
================

Provides Raman spectrum baseline removal tools.
"""
from typing import Tuple

import numpy as np
from numba import njit
from scipy.signal import savgol_filter

# imodpoly


def imodpoly(
    spectrum: np.ndarray,
    poly_order: int = 6,
    precision: float = 0.005,
    max_iter: int = 1000,
    imod: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    imodpoly splits a spectrum into it's raman and baseline components.
    Based on the IModPoly (or ModPoly if using imod=False) algorithm.

    Usage
    ------
    (raman, baseline) = imodpoly(spectrum,
                                 poly_order=6,
                                 precision=0.005,
                                 max_iter=1000,
                                 imod=True
                                 )

    Parameters
    ----------
    spectrum : np.ndarray
        the input spectrum
    poly_order : int, optional
        the polinomial fit's order, by default 6
    precision : float, optional
        the precision to reach. , by default 0.005
    max_iter : int, optional
        maximum number of iterations, by default 1000
    imod : bool, optional
        imod=True will use imodpoly
        imod=False will use modpoly, by default True

    Returns
    -------
    raman : np.ndarray
        the spectrum's raman component
    baseline : np.ndarray
        the spectrum's baseline component

    Reference
    ---------
    'Automated Autofluorescence Background Subtraction Algorithm for
    Biomedical Raman Spectroscopy'
    DOI : 10.1366/000370207782597003
    """

    i = 1
    converged = False
    raman_data = np.array(spectrum)
    xaxis = np.array(range(0, raman_data.shape[0]))
    std_dev = 0
    while not converged and i < max_iter:
        poly_fit = np.polyval(np.polyfit(xaxis, raman_data, poly_order), xaxis)
        residual = raman_data - poly_fit
        previous_std_dev = std_dev
        std_dev = np.std(residual)

        # if first iteration -> peak removal
        if imod:
            # IModPoly
            if i == 1:
                ind = np.where(raman_data > poly_fit + std_dev)[0]
                raman_data[ind] = poly_fit[ind] + std_dev

            ind = np.where(raman_data > poly_fit + std_dev)[0]
            raman_data[ind] = poly_fit[ind] + std_dev

        else:
            # ModPoly
            if i == 1:
                ind = np.where(raman_data > poly_fit)[0]
                raman_data[ind] = poly_fit[ind]

            ind = np.where(raman_data > poly_fit)[0]
            raman_data[ind] = poly_fit[ind]

        converged = np.abs((std_dev - previous_std_dev) / std_dev) < precision
        i = i + 1

    baseline = poly_fit
    raman = spectrum - baseline
    return raman, baseline


# Morphological transformations


@njit(cache=True)
def erosion(signal: list, hws: list) -> list:
    """
    erosion computes the morphological erosion of a singal using
    a plane structuring element window.

    Usage
    -----
    signal_ = erosion(signal, hws)

    Parameters
    ----------
    signal : list
        the input signal
    hws : list
        the Half Window Size

    Returns
    -------
    signal_ : list
        the eroded signal
    """
    eroded_f = []
    for i in range(len(signal)):
        left_bound = i - hws[i]  # left bound of window
        if left_bound < 0:
            left_bound = 0
        right_bound = i + hws[i] + 1  # right bound of window
        if right_bound > len(signal):
            right_bound = len(signal)
        eroded_f.append(min(signal[left_bound:right_bound]))
    return eroded_f


@njit(cache=True)
def dilation(signal: list, hws: list) -> list:
    """
    dilation computes the morphological dilation of a signal using
    a plane structuring element window.

    Usage
    -----
    signal_ = dilation(signal, hws)

    Parameters
    ----------
    signal : list
        the input signal
    hws : list
        the Half Window Size

    Returns
    -------
    signal_ : list
        the dilated signal
    """

    diladed_f = []
    for i in range(len(signal)):
        lbound = i - hws[i]  # left bound of window
        if lbound < 0:
            lbound = 0
        rbound = i + hws[i] + 1  # right bound of window
        if rbound > len(signal):
            rbound = len(signal)
        diladed_f.append(max(signal[lbound:rbound]))
    return diladed_f


@njit(cache=True)
def opening(signal: list, hws: list) -> list:
    """
    opening computes the morphological opening of a signal using
    a plane structuring element window.

    Usage
    -----
    signal_ = opening(signal, hws)

    Parameters
    ----------
    signal : list
        the input signal
    hws : list
        the Half Window Size

    Returns
    -------
    signal_ : list
        the opened signal

    Notes
    -----
    The morphological opening is the same as an erosion followed
    by a dilation : opening{X} = dilation{erosion{X}}
    """

    eroded_f = erosion(signal, hws)
    opened_f = dilation(eroded_f, hws)
    return opened_f


@njit(cache=True)
def bopening(signal: list, hws: list) -> list:
    """
    bopening computes the better opening (see reference) of a signal using
    a structuring element window.

    Usage
    -----
    signal_ = opening(signal, hws)

    Parameters
    ----------
    signal : list
        the input signal
    hws : list
        the the Half Window Size

    Returns
    -------
    signal_ : list
        the b-opened signal

    Reference
    ---------
    'Morphology-Based Automated Baseline Removal for Raman
    Spectra of Artistic Pigments'
    DOI:10.1366/000370210791414281
    """

    opened_f = opening(signal, hws)  # \gamma(f)
    dilated_opening = dilation(opened_f, hws)
    eroded_opening = erosion(opened_f, hws)
    opened_mod = [
        (dilated_opening[i] + eroded_opening[i]) / 2 for i in range(len(signal))
    ]  # \gamma'(f)
    bopened = [min([opened_mod[i], opened_f[i]]) for i in range(len(signal))]
    return bopened


# Morphology based baseline removal


def morph_br(spectrum: np.ndarray, hws: int) -> np.ndarray:
    """
    morph_br splits a spectrum into it's raman and baseline components

    Usage
    -----
    raman, baseline = morph_br(spectrum, hws)

    Parameters
    ----------
    spectrum : np.ndarray
        the input spectrum
    hws : int
        the Half Window Size

    Returns
    -------
    raman : np.ndarray
        the spectrum's raman component
    baseline : np.ndarray
        the spectrum's baseline component

    Reference
    ---------
    'Morphology-Based Automated Baseline Removal for Raman
    Spectra of Artistic Pigments'
    DOI:10.1366/000370210791414281
    """
    if isinstance(hws, int):
        if hws < 1:
            raise ValueError("Minimal hws is 1")
        hws = hws * np.ones(len(spectrum))
        hws = hws.astype(int)
    elif isinstance(hws, list):
        hws = np.array(hws).astype(int)
    elif isinstance(hws, np.ndarray):
        hws = hws.astype(int)
    else:
        raise TypeError("hws must be one of the following ; INT, LIST or NDARRAY")

    spectrum = np.array(spectrum)
    baseline = np.array(bopening(spectrum, hws))
    raman = np.array([spectrum[i] - baseline[i] for i in range(len(spectrum))])
    return raman, baseline


# Bubblefill


@njit(cache=True)
def grow_bubble(
    spectrum: np.ndarray, alignment: str = "center"
) -> Tuple[np.ndarray, int]:
    """
    grow_bubble grows a bubble until it touches a spectrum.

    Usage
    -----
    bubble, touching_point = grow_bubble(spectrum, alignment)

    Parameters
    ----------
    spectrum : np.ndarray
        the input spectrum
    alignment : str, optional
        the alignment of the bubble to be grown. Possible values are
        ['left', 'right' and 'center'].
        If 'left', the bubble will be centered in x=0 and will have a width
        of 2x len(spectrum)
        If 'right', the bubble will be centered in x=len(spectrum) and will
        have a width of 2x len(spectrum)
        If 'center', the bubble will be centered in x=len(spectrum)/2 and
        will have a width of len(spectrum)
        , by default "center"

    Returns
    -------
    bubble : np.ndarray
        the grown bubble
    touching_point : int
        the x-coordinate where it touched the spectrum and *popped*.
    """
    xaxis = np.arange(len(spectrum))

    # Ajusting bubble parameter based on alignment
    if alignment == "left":
        # half bubble right
        width = 2 * len(spectrum)
        middle = 0
    elif alignment == "right":
        # half bubble left
        width = 2 * len(spectrum)
        middle = len(spectrum)
    else:
        # Centered bubble
        width = len(spectrum)
        middle = len(spectrum) / 2

    squared_arc = (width / 2) ** 2 - (xaxis - middle) ** 2  # squared half circle
    # squared_arc[squared_arc < 0] = 0
    bubble = np.sqrt(squared_arc) - width
    # find new intersection
    touching_point = (spectrum - bubble).argmin()

    # grow bubble until touching
    bubble = bubble + (spectrum - bubble).min()

    return bubble, touching_point


@njit(cache=True)
def keep_largest(baseline: np.ndarray, bubble: np.ndarray) -> np.ndarray:
    """
    keep_largest selectively updates a baseline region with a bubble, depending
    on which has a greater y-value.

    Usage
    -----
    baseline_ = keep_largest(baseline, bubble)

    Parameters
    ----------
    baseline : np.ndarray
        an input baseline
    bubble : np.ndarray
        an input bubble (usually computed with grow_bubble())

    Returns
    -------
    baseline_ : np.ndarray
        the updated baseline
    """
    for i, _ in enumerate(baseline):
        if baseline[i] < bubble[i]:
            baseline[i] = bubble[i]
    return baseline


def bubbleloop(
    spectrum: np.ndarray, baseline: np.ndarray, min_bubble_widths: list
) -> np.ndarray:
    """
    bubbleloop itteratively updates a baseline estimate by growing bubbles under a spectrum.

    Usage
    -----
    baseline = bubbleloop(spectrum, baseline, min_bubble_widths)

    Parameters
    ----------
    spectrum : np.ndarray
        the input spectrum
    baseline : np.ndarray
        the initial baseline should be akin to np.zeros(spectrum.shape)
    min_bubble_widths : list
        the minimum bubble widths to use. Can be an array-like or int.
        if array-like -> must be the same length as spectrum and baseline. Useful to specify
        different bubble sizes based on x-coordinates.
        if int -> will use the same width for all x-coordinates.

    Returns
    -------
    baseline : np.ndarray
        the updated baseline
    """
    # initial range is always 0 -> len(s). aka the whole spectrum
    # bubblecue is a list of bubble x-coordinate span as
    # [[x0, x2]_0, [x0, x2]_1, ... [x0, x2]_n]
    # additional bubble regions are added as the loop runs.
    range_cue = [[0, len(spectrum)]]

    i = 0
    while i < len(range_cue):
        # Bubble parameter from bubblecue
        left_bound, right_bound = range_cue[i]
        i += 1

        if left_bound == right_bound:
            continue

        if isinstance(min_bubble_widths, int):
            min_bubble_width = min_bubble_widths
        else:
            min_bubble_width = min_bubble_widths[(left_bound + right_bound) // 2]

        if left_bound == 0 and right_bound != (len(spectrum)):
            # half bubble right
            alignment = "left"
        elif left_bound != 0 and right_bound == (len(spectrum)):
            alignment = "right"
            # half bubble left
        else:

            # Reached minimum bubble width
            if (right_bound - left_bound) < min_bubble_width:
                continue
            # centered bubble
            alignment = "center"

        # new bubble
        bubble, relative_touching_point = grow_bubble(
            spectrum[left_bound:right_bound], alignment
        )
        touching_point = relative_touching_point + left_bound

        # add bubble to baseline by keeping largest value
        baseline[left_bound:right_bound] = keep_largest(
            baseline[left_bound:right_bound], bubble
        )
        # Add new bubble(s) to bubblecue
        if touching_point == left_bound:
            range_cue.append([touching_point + 1, right_bound])
        elif touching_point == right_bound:
            range_cue.append([left_bound, touching_point - 1])
        else:
            range_cue.append([left_bound, touching_point])
            range_cue.append([touching_point, right_bound])

    return baseline


def bubblefill(
    spectrum: np.ndarray, min_bubble_widths: list = 50, fit_order: int = 1
) -> Tuple[np.ndarray, np.ndarray]:
    """
    bubblefill splits a spectrum into it's raman and baseline components.

    Usage
    -----
    raman, baseline = bubblefill(spectrum, bubblewidths, fitorder)

    Parameters
    ----------
    spectrum : np.ndarray
        the input spectrum
    min_bubble_widths: list or int
        is the smallest width allowed for bubbles. Smaller values will
        allow bubbles to penetrate further into peaks resulting
        in a more *aggressive* baseline removal. Larger values are more
        *concervative* and might the computed underestimate baseline.
        use list to specify a minimum width that depends on the
        x-coordinate of the bubble. Make sure len(bubblewidths) = len(spectrum).
        Otherwise if bubblewidths [int], the same width is used for all x-coordinates.
    fit_order : int
        the order of the polynomial fit used to remove the *overall* baseline slope.
        Recommendend value is 1 (for linear slope).
        Higher order will result in Runge's phenomena and
        potentially undesirable and unpredictable effects.
        fitorder = 0 is the same as not removing the overall baseline slope

    Returns
    -------
    raman : np.ndarray
        the spectrum's raman component
    baseline : np.ndarray
        the spectrum's baseline component

    Reference
    ---------
    Guillaume Sheehy 2021-01
    """
    xaxis = np.arange(len(spectrum))

    # Remove general slope
    poly_fit = np.poly1d(np.polyfit(xaxis, spectrum, fit_order))(xaxis)
    spectrum_ = spectrum - poly_fit

    # Normalization
    smin = spectrum_.min()  # value needed to return to the original scaling
    spectrum_ = spectrum_ - smin
    scale = spectrum_.max() / len(spectrum)
    spectrum_ = spectrum_ / scale  # Rescale spectrum to X:Y=1:1 (square aspect ratio)

    baseline = np.zeros(spectrum_.shape)

    # Bubble loop (this is the bulk of the algorithm)
    baseline = bubbleloop(spectrum_, baseline, min_bubble_widths)

    # Bringing baseline back in original scale
    baseline = baseline * scale + poly_fit + smin

    # Final smoothing of baseline (only if bubblewidth is not a list!!!)
    if not isinstance(min_bubble_widths, int):
        filter_width = max(min(min_bubble_widths), 10)
    else:
        filter_width = max(min_bubble_widths, 10)
    # print(filter_width)
    baseline = savgol_filter(baseline, int(2 * (filter_width // 4) + 3), 3)

    raman = spectrum - baseline

    return raman, baseline
